"""Isotope scattering calculation."""
# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from phonopy.harmonic.dynamical_matrix import get_dynamical_matrix
from phonopy.phonon.tetrahedron_mesh import get_tetrahedra_frequencies
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.atoms import isotope_data as phonopy_isotope_data
from phonopy.structure.cells import Primitive
from phonopy.structure.symmetry import Symmetry
from phonopy.structure.tetrahedron_method import TetrahedronMethod
from phonopy.units import VaspToTHz

from phono3py.other.tetrahedron_method import (
    get_integration_weights,
    get_unique_grid_points,
)
from phono3py.phonon.func import gaussian
from phono3py.phonon.grid import BZGrid
from phono3py.phonon.solver import run_phonon_solver_c, run_phonon_solver_py


def get_mass_variances(
    primitive: Optional[PhonopyAtoms] = None,
    symbols: Optional[Union[List[str], Tuple[str]]] = None,
    isotope_data: Optional[Dict] = None,
):
    """Calculate mass variances."""
    if primitive is not None:
        _symbols = primitive.symbols
    elif symbols is not None:
        _symbols = symbols
    else:
        raise RuntimeError("primitive or symbols have to be given.")

    _isotope_data = {}
    for s in _symbols:
        if isotope_data is not None and s in isotope_data:
            _isotope_data[s] = isotope_data[s]
        else:
            _isotope_data[s] = phonopy_isotope_data[s]

    mass_variances = []
    for s in _symbols:
        masses = np.array([x[1] for x in _isotope_data[s]])
        fractions = np.array([x[2] for x in _isotope_data[s]])
        m_ave = np.dot(masses, fractions)
        g = np.dot(fractions, (1 - masses / m_ave) ** 2)
        mass_variances.append(g)

    return np.array(mass_variances, dtype="double")


class Isotope:
    """Isotope scattering calculation class."""

    def __init__(
        self,
        mesh,
        primitive,
        mass_variances=None,  # length of list is num_atom.
        isotope_data=None,
        band_indices=None,
        sigma=None,
        bz_grid=None,
        frequency_factor_to_THz=VaspToTHz,
        use_grg=False,
        store_dense_gp_map=True,
        symprec=1e-5,
        cutoff_frequency=None,
        lapack_zheev_uplo="L",
    ):
        """Init method."""
        self._mesh = mesh
        if mass_variances is None:
            self._mass_variances = get_mass_variances(
                primitive, isotope_data=isotope_data
            )
        else:
            self._mass_variances = np.array(mass_variances, dtype="double")
        self._primitive = primitive
        self._sigma = sigma
        self._bz_grid = bz_grid
        self._symprec = symprec
        if cutoff_frequency is None:
            self._cutoff_frequency = 0
        else:
            self._cutoff_frequency = cutoff_frequency
        self._frequency_factor_to_THz = frequency_factor_to_THz
        self._lapack_zheev_uplo = lapack_zheev_uplo
        self._nac_q_direction = None

        self._grid_points = None
        self._frequencies = None
        self._eigenvectors = None
        self._phonon_done = None
        self._dm = None
        self._gamma = None
        self._tetrahedron_method = None

        num_band = len(self._primitive) * 3
        if band_indices is None:
            self._band_indices = np.arange(num_band, dtype="int_")
        else:
            self._band_indices = np.array(band_indices, dtype="int_")

        if self._bz_grid is None:
            primitive_symmetry = Symmetry(self._primitive, self._symprec)
            self._bz_grid = BZGrid(
                self._mesh,
                lattice=self._primitive.cell,
                symmetry_dataset=primitive_symmetry.dataset,
                use_grg=use_grg,
                store_dense_gp_map=store_dense_gp_map,
            )

    def set_grid_point(self, grid_point):
        """Initialize grid points."""
        self._grid_point = grid_point
        self._grid_points = np.arange(len(self._bz_grid.addresses), dtype="int_")

        if self._phonon_done is None:
            self._allocate_phonon()

    def run(self, lang="C"):
        """Run isotope scattering calculation."""
        if lang == "C":
            self._run_c()
        else:
            self._run_py()

    @property
    def sigma(self):
        """Setter and getter of smearing width."""
        return self._sigma

    @sigma.setter
    def sigma(self, sigma):
        if sigma is None:
            self._sigma = None
        else:
            self._sigma = float(sigma)

    def set_sigma(self, sigma):
        """Set smearing width."""
        warnings.warn(
            "Isotope.set_sigma() is deprecated." "Use Isotope.sigma attribute.",
            DeprecationWarning,
        )
        self.sigma = sigma

    @property
    def dynamical_matrix(self):
        """Return DynamicalMatrix* class instance."""
        return self._dm

    @property
    def band_indices(self):
        """Return specified band indices."""
        return self._band_indices

    @property
    def gamma(self):
        """Return scattering strength."""
        return self._gamma

    def get_gamma(self):
        """Return scattering strength."""
        warnings.warn(
            "Isotope.get_gamma() is deprecated." "Use Isotope.gamma attribute.",
            DeprecationWarning,
        )
        return self.gamma

    @property
    def bz_grid(self):
        """Return BZgrid class instance."""
        return self._bz_grid

    @property
    def mass_variances(self):
        """Return mass variances."""
        return self._mass_variances

    def get_mass_variances(self):
        """Return mass variances."""
        warnings.warn(
            "Isotope.get_mass_variances() is deprecated."
            "Use Isotope.mass_variances attribute.",
            DeprecationWarning,
        )
        return self.mass_variances

    def get_phonons(self):
        """Return phonons on grid."""
        return self._frequencies, self._eigenvectors, self._phonon_done

    def set_phonons(self, frequencies, eigenvectors, phonon_done, dm=None):
        """Set phonons on grid."""
        self._frequencies = frequencies
        self._eigenvectors = eigenvectors
        self._phonon_done = phonon_done
        if dm is not None:
            self._dm = dm

    def init_dynamical_matrix(
        self,
        fc2,
        supercell: PhonopyAtoms,
        primitive: Primitive,
        nac_params=None,
        frequency_scale_factor=None,
        decimals=None,
    ):
        """Initialize dynamical matrix."""
        self._primitive = primitive
        self._dm = get_dynamical_matrix(
            fc2,
            supercell,
            primitive,
            nac_params=nac_params,
            frequency_scale_factor=frequency_scale_factor,
            decimals=decimals,
            symprec=self._symprec,
        )

    def set_nac_q_direction(self, nac_q_direction=None):
        """Set q-direction at q->0 used for NAC."""
        if nac_q_direction is not None:
            self._nac_q_direction = np.array(nac_q_direction, dtype="double")

    def _run_c(self):
        self._run_phonon_solver_c(self._grid_points)
        import phono3py._phono3py as phono3c

        gamma = np.zeros(len(self._band_indices), dtype="double")
        weights_in_bzgp = np.ones(len(self._grid_points), dtype="double")
        if self._sigma is None:
            self._set_integration_weights()
            phono3c.thm_isotope_strength(
                gamma,
                self._grid_point,
                self._bz_grid.grg2bzg,
                weights_in_bzgp,
                self._mass_variances,
                self._frequencies,
                self._eigenvectors,
                self._band_indices,
                self._integration_weights,
                self._cutoff_frequency,
            )
        else:
            phono3c.isotope_strength(
                gamma,
                self._grid_point,
                self._bz_grid.grg2bzg,
                weights_in_bzgp,
                self._mass_variances,
                self._frequencies,
                self._eigenvectors,
                self._band_indices,
                self._sigma,
                self._cutoff_frequency,
            )

        self._gamma = gamma / np.prod(self._bz_grid.D_diag)

    def _set_integration_weights(self, lang="C"):
        if lang == "C":
            self._set_integration_weights_c()
        else:
            self._set_integration_weights_py()

    def _set_integration_weights_c(self):
        """Set tetrahedron method integration weights.

        self._frequencies are those on all BZ-grid. So all those grid points in
        BZ-grid, i.e., self._grid_points, are passed to get_integration_weights.

        """
        unique_grid_points = get_unique_grid_points(self._grid_points, self._bz_grid)
        self._run_phonon_solver_c(unique_grid_points)
        freq_points = np.array(
            self._frequencies[self._grid_point, self._band_indices],
            dtype="double",
            order="C",
        )
        self._integration_weights = get_integration_weights(
            freq_points, self._frequencies, self._bz_grid, grid_points=self._grid_points
        )

    def _set_integration_weights_py(self):
        """Set tetrahedron method integration weights.

        Python implementation corresponding to _set_integration_weights_c.

        """
        thm = TetrahedronMethod(self._bz_grid.microzone_lattice)
        num_grid_points = len(self._grid_points)
        num_band = len(self._primitive) * 3
        self._integration_weights = np.zeros(
            (num_grid_points, len(self._band_indices), num_band), dtype="double"
        )

        for i, gp in enumerate(self._grid_points):
            tfreqs = get_tetrahedra_frequencies(
                gp,  # In BZ-grid used only to access self._bz_grid.addresses.
                self._bz_grid.D_diag,
                self._bz_grid.addresses,
                np.array(
                    np.dot(thm.tetrahedra, self._bz_grid.P.T),
                    dtype="int_",
                    order="C",
                ),
                self._bz_grid.grg2bzg,
                self._frequencies,
                grid_order=[
                    1,
                    self._bz_grid.D_diag[0],
                    self._bz_grid.D_diag[0] * self._bz_grid.D_diag[1],
                ],
                lang="Py",
            )

            for bi, frequencies in enumerate(tfreqs):
                thm.set_tetrahedra_omegas(frequencies)
                thm.run(self._frequencies[self._grid_point, self._band_indices])
                iw = thm.get_integration_weight()
                self._integration_weights[i, :, bi] = iw

    def _run_py(self):
        for gp in self._grid_points:
            self._run_phonon_solver_py(gp)

        if self._sigma is None:
            self._set_integration_weights(lang="Py")

        t_inv = []
        for bi in self._band_indices:
            vec0 = self._eigenvectors[self._grid_point][:, bi].conj()
            f0 = self._frequencies[self._grid_point][bi]
            ti_sum = 0.0
            for gp in self._bz_grid.grg2bzg:
                for j, (f, vec) in enumerate(
                    zip(self._frequencies[gp], self._eigenvectors[gp].T)
                ):
                    if f < self._cutoff_frequency:
                        continue
                    ti_sum_band = np.sum(
                        np.abs((vec * vec0).reshape(-1, 3).sum(axis=1)) ** 2
                        * self._mass_variances
                    )
                    if self._sigma is None:
                        ti_sum += ti_sum_band * self._integration_weights[gp, bi, j]
                    else:
                        ti_sum += ti_sum_band * gaussian(f0 - f, self._sigma)
            t_inv.append(np.pi / 2 / np.prod(self._bz_grid.D_diag) * f0**2 * ti_sum)

        self._gamma = np.array(t_inv, dtype="double") / 2

    def _run_phonon_solver_c(self, grid_points):
        run_phonon_solver_c(
            self._dm,
            self._frequencies,
            self._eigenvectors,
            self._phonon_done,
            grid_points,
            self._bz_grid.addresses,
            self._bz_grid.QDinv,
            self._frequency_factor_to_THz,
            self._nac_q_direction,
            self._lapack_zheev_uplo,
        )

    def _run_phonon_solver_py(self, grid_point):
        run_phonon_solver_py(
            grid_point,
            self._phonon_done,
            self._frequencies,
            self._eigenvectors,
            self._bz_grid.addresses,
            self._bz_grid.QDinv,
            self._dm,
            self._frequency_factor_to_THz,
            self._lapack_zheev_uplo,
        )

    def _allocate_phonon(self):
        num_band = len(self._primitive) * 3
        num_grid = len(self._bz_grid.addresses)
        self._phonon_done = np.zeros(num_grid, dtype="byte")
        self._frequencies = np.zeros((num_grid, num_band), dtype="double")
        itemsize = self._frequencies.itemsize
        self._eigenvectors = np.zeros(
            (num_grid, num_band, num_band), dtype=("c%d" % (itemsize * 2))
        )



"""Calculate group velocity matrix."""
# Copyright (C) 2021 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from phonopy.phonon.thermal_properties import mode_cv
from phonopy.units import Kb


def mode_cv_matrix(temp, freqs, cutoff=1e-4):
    r"""Calculate mode heat capacity matrix, Cqjj'.

    C_{\mathbf{q}jj'} = K_B \frac{e^{x_{\mathbf{q}j} - x_{\mathbf{q}j'}} - 1}
    {x_{\mathbf{q}j} - x_{\mathbf{q}j'}} \left( \frac{
    x_{\mathbf{q}j} + x_{\mathbf{q}j'}}{2}
    \right)^2 n_{\mathbf{q}j}(n_{\mathbf{q}j'} + 1)

    Note
    ----
    Diagonal (j=j') terms reduce to normal mode heat capacity.

    Parameters
    ----------
    temp : float
        Temperature in K.
    freqs : ndarray
        Phonon frequencies at a q-point in eV.
    cutoff : float
        This is used to check the degeneracy.

    Returns
    -------
    ndarray
        Heat capacity matrix in eV/K.
        shape=(num_band, num_band), dtype='double', order='C'.

    """
    x = freqs / Kb / temp
    shape = (len(freqs), len(freqs))
    cvm = np.zeros(shape, dtype="double", order="C")
    for i, j in np.ndindex(shape):
        if abs(freqs[i] - freqs[j]) < cutoff:
            cvm[i, j] = mode_cv(temp, freqs[i])
            continue
        sub = x[i] - x[j]
        add = x[i] + x[j]
        n_inv = np.exp([x[i], x[j], sub]) - 1
        cvm[i, j] = Kb * n_inv[2] / sub * (add / 2) ** 2 / n_inv[0] * (1 / n_inv[1] + 1)
    return cvm



"""Thermal conductivity base class."""
# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import textwrap
import warnings
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

import numpy as np
from phonopy.phonon.group_velocity import GroupVelocity
from phonopy.phonon.thermal_properties import mode_cv
from phonopy.units import EV, Angstrom, Kb, THz, THzToEv

from phono3py.other.isotope import Isotope
from phono3py.phonon3.collision_matrix import CollisionMatrix
from phono3py.phonon3.imag_self_energy import ImagSelfEnergy
from phono3py.phonon3.interaction import Interaction
from phono3py.phonon.grid import get_grid_points_by_rotations, get_ir_grid_points

unit_to_WmK = (
    (THz * Angstrom) ** 2 / (Angstrom**3) * EV / THz / (2 * np.pi)
)  # 2pi comes from definition of lifetime.


class HeatCapacityMixIn:
    """Heat capacity mix-in.

    Used by other mix-in.

    """

    @property
    def mode_heat_capacities(self):
        """Return mode heat capacity at constant volume at grid points.

        Grid points are those at mode kappa are calculated.

        """
        return self._cv

    def get_mode_heat_capacities(self):
        """Return mode heat capacity at constant volume at grid points.

        Grid points are those at mode kappa are calculated.

        """
        warnings.warn(
            "Use attribute, Conductivity.mode_heat_capacities "
            "instead of Conductivity.get_mode_heat_capacities().",
            DeprecationWarning,
        )
        return self.mode_heat_capacities

    def _set_cv(self, i_gp, i_data):
        """Set mode heat capacity.

        The array has to be allocated somewhere out of the mix-in.

        self._cv = np.zeros(
            (num_temp, num_grid_points, num_band0), order="C", dtype="double"
        )

        """
        grid_point = self._grid_points[i_gp]
        freqs = self._frequencies[grid_point][self._pp.band_indices] * THzToEv
        cutoff = self._pp.cutoff_frequency * THzToEv
        cv = np.zeros((len(self._temperatures), len(freqs)), dtype="double")
        # x=freq/T has to be small enough to avoid overflow of exp(x).
        # x < 100 is the hard-corded criterion.
        # Otherwise just set 0.
        for i, f in enumerate(freqs):
            if f > cutoff:
                condition = f < 100 * self._temperatures * Kb
                cv[:, i] = np.where(
                    condition,
                    mode_cv(np.where(condition, self._temperatures, 10000), f),
                    0,
                )
        self._cv[:, i_data, :] = cv


class ConductivityMixIn(HeatCapacityMixIn):
    """Thermal conductivity mix-in.

    Used by ConductivityRTA and ConductivityLBTE.

    """

    @property
    def kappa(self):
        """Return kappa."""
        return self._kappa

    def get_kappa(self):
        """Return kappa."""
        warnings.warn(
            "Use attribute, Conductivity.kappa " "instead of Conductivity.get_kappa().",
            DeprecationWarning,
        )
        return self.kappa

    @property
    def mode_kappa(self):
        """Return mode_kappa."""
        return self._mode_kappa

    def get_mode_kappa(self):
        """Return mode_kappa."""
        warnings.warn(
            "Use attribute, Conductivity.mode_kappa "
            "instead of Conductivity.get_mode_kappa().",
            DeprecationWarning,
        )
        return self.mode_kappa

    @property
    def group_velocities(self):
        """Return group velocities at grid points.

        Grid points are those at mode kappa are calculated.

        """
        return self._gv

    def get_group_velocities(self):
        """Return group velocities at grid points.

        Grid points are those at mode kappa are calculated.

        """
        warnings.warn(
            "Use attribute, Conductivity.group_velocities "
            "instead of Conductivity.get_group_velocities().",
            DeprecationWarning,
        )
        return self.group_velocities

    @property
    def gv_by_gv(self):
        """Return gv_by_gv at grid points where mode kappa are calculated."""
        return self._gv_sum2

    def get_gv_by_gv(self):
        """Return gv_by_gv at grid points where mode kappa are calculated."""
        warnings.warn(
            "Use attribute, Conductivity.gv_by_gv "
            "instead of Conductivity.get_gv_by_gv().",
            DeprecationWarning,
        )
        return self.gv_by_gv

    def _init_velocity(self, gv_delta_q):
        self._velocity_obj = GroupVelocity(
            self._pp.dynamical_matrix,
            q_length=gv_delta_q,
            symmetry=self._pp.primitive_symmetry,
            frequency_factor_to_THz=self._pp.frequency_factor_to_THz,
        )

    def _set_velocities(self, i_gp, i_data):
        self._gv[i_data] = self._get_gv(i_gp)
        self._set_gv_by_gv(i_gp, i_data)

    def _get_gv(self, i_gp):
        """Get group velocity."""
        irgp = self._grid_points[i_gp]

        if self._average_gv_over_kstar and len(self._point_operations) > 1:
            gps_rotated = get_grid_points_by_rotations(
                irgp, self._pp.bz_grid, with_surface=True
            )
            assert len(gps_rotated) == len(self._point_operations)

            unique_gps = np.unique(gps_rotated)
            gvs = {}
            for bz_gp in unique_gps.tolist():  # To conver to int type.
                self._velocity_obj.run([self._get_qpoint_from_gp_index(bz_gp)])
                gvs[bz_gp] = self._velocity_obj.group_velocities[
                    0, self._pp.band_indices, :
                ]
            gv = np.zeros_like(gvs[irgp])
            for bz_gp, r in zip(gps_rotated, self._rotations_cartesian):
                gv += np.dot(gvs[bz_gp], r)  # = dot(r_inv, gv)
            return gv / len(self._point_operations)
        else:
            self._velocity_obj.run([self._get_qpoint_from_gp_index(irgp)])
            return self._velocity_obj.group_velocities[0, self._pp.band_indices, :]

    def _set_gv_by_gv(self, i_gp, i_data):
        """Outer product of group velocities.

        (v x v) [num_k*, num_freqs, 3, 3]

        """
        gv_by_gv_tensor, order_kstar = self._get_gv_by_gv(i_gp, i_data)
        self._num_sampling_grid_points += order_kstar

        # Sum all vxv at k*
        for j, vxv in enumerate(([0, 0], [1, 1], [2, 2], [1, 2], [0, 2], [0, 1])):
            self._gv_sum2[i_data, :, j] = gv_by_gv_tensor[:, vxv[0], vxv[1]]

    def _get_gv_by_gv(self, i_gp, i_data):
        multi = self._get_multiplicity_at_q(i_gp)
        gv = self._gv[i_data]
        gv_by_gv = np.zeros((len(gv), 3, 3), dtype="double")
        for r in self._rotations_cartesian:
            gvs_rot = np.dot(gv, r.T)
            gv_by_gv += [np.outer(r_gv, r_gv) for r_gv in gvs_rot]
        gv_by_gv /= multi
        return gv_by_gv, self._get_kstar_order(i_gp, multi)


class ConductivityBase(ABC):
    """Base class of Conductivity classes.

    All Conductivity* classes have to inherit this base class.

    self._gv has to be allocated in the inherited classes.

    """

    _average_gv_over_kstar = False

    def __init__(
        self,
        interaction: Interaction,
        grid_points=None,
        temperatures: Optional[Union[List, np.ndarray]] = None,
        sigmas: Optional[Union[List, np.ndarray]] = None,
        sigma_cutoff: Optional[float] = None,
        is_isotope=False,
        mass_variances: Optional[Union[List, np.ndarray]] = None,
        boundary_mfp: Optional[float] = None,
        is_kappa_star=True,
        gv_delta_q=None,
        is_full_pp=False,
        log_level=0,
    ):
        """Init method.

        interaction : Interaction
            Interaction class instance.
        grid_points : array_like or None, optional
            Grid point indices in BZgrid. When None, ir-grid points are searched
            internally. Default is None.
            shape=(grid_points, ), dtype='int_'.
        temperatures : array_like, optional, default is None
            Temperatures at which thermal conductivity is calculated.
            shape=(temperature_points, ), dtype='double'.
        sigmas : array_like, optional, default is None
            The float values are given as the standard deviations of Gaussian
            function. If None is given as an element of this list, linear
            tetrahedron method is used instead of smearing method.
        sigma_cutoff : float, optional, default is None
            This is given as a multiple of the standard deviation. For example,
            if this value is 5, the tail of the Gaussian function is cut at 5 sigma.
        is_isotope : bool, optional, default is False
            With or without isotope scattering.
        mass_variances : array_like, optional, default is None
            Mass variances for isotope scattering calculation. When None,
            the values stored in phono3py are used with `is_isotope=True`.
            shape(atoms_in_primitive, ), dtype='double'.
        boundary_mfp : float, optiona, default is None
            Mean free path in micrometre to calculate simple boundary
            scattering contribution to thermal conductivity.
            None gives 1 metre, which is supposed negligible contribution,
            but this value is used to avoid divergence of phonon lifetime.
        is_kappa_star : bool, optional
            When True, reciprocal space symmetry is used to calculate
            lattice thermal conductivity. This calculation is peformed
            iterating over specific grid points. With `is_kappa_star=True`
            and `grid_points=None`, ir-grid points are used for the iteration.
            Default is True.
        gv_delta_q : float, optional, default is None,  # for group velocity
            With non-analytical correction, group velocity is calculated
            by central finite difference method. This value gives the distance
            in both directions in 1/Angstrom. The default value will be 1e-5.
        is_full_pp : bool, optional, default is False
            With True, full elements of phonon-phonon interaction strength
            are computed. However with tetrahedron method, part of them are
            known to be zero and unnecessary to calculation. With False,
            those elements are not calculated, by which considerable
            improve of efficiency is expected.
            With smearing method, even if this is set False, full elements
            are computed unless `sigma_cutoff` is specified.
        log_level : int, optional
            Verbosity control. Default is 0.

        """
        self._pp: Interaction = interaction
        self._is_kappa_star = is_kappa_star
        self._is_full_pp = is_full_pp
        self._log_level = log_level
        self._complex_dtype = "c%d" % (np.dtype("double").itemsize * 2)

        self._point_operations, self._rotations_cartesian = self._get_point_operations()
        (
            self._grid_points,
            self._ir_grid_points,
            self._grid_weights,
        ) = self._get_grid_info(grid_points)
        self._grid_point_count: int = 0
        self._num_sampling_grid_points: int = 0

        self._sigmas: List
        if sigmas is None:
            self._sigmas = []
        else:
            self._sigmas = list(sigmas)
        self._sigma_cutoff = sigma_cutoff
        self._collision: Union[ImagSelfEnergy, CollisionMatrix]
        self._temperatures: Optional[np.ndarray]
        if temperatures is None:
            self._temperatures = None
        else:
            self._temperatures = np.array(temperatures, dtype="double")
        self._boundary_mfp = boundary_mfp

        self._pp.nac_q_direction = None
        (
            self._frequencies,
            self._eigenvectors,
            self._phonon_done,
        ) = self._pp.get_phonons()
        if (self._phonon_done == 0).any():
            self._pp.run_phonon_solver()

        self._is_isotope = is_isotope
        if mass_variances is not None:
            self._is_isotope = True
        self._isotope: Isotope
        self._mass_variances: np.ndarray
        if self._is_isotope:
            self._set_isotope(mass_variances)

        self._read_gamma = False
        self._read_gamma_iso = False

        # Allocated in self._allocate_values.
        self._gv: np.ndarray
        self._gamma: np.ndarray
        self._gamma_iso: Optional[np.ndarray] = None

        volume = self._pp.primitive.volume
        self._conversion_factor = unit_to_WmK / volume

        # `self._velocity_obj` is the instance of an inherited class of
        # `GroupVelocity`. `self._init_velocity()` is the method setup the instance,
        # which must be implmented in the inherited class of `ConductivityBase`.
        self._velocity_obj: GroupVelocity
        self._init_velocity(gv_delta_q)

    def __iter__(self):
        """Calculate mode kappa at each grid point."""
        return self

    def __next__(self):
        """Return grid point count for mode kappa."""
        if self._grid_point_count == len(self._grid_points):
            if self._log_level:
                print(
                    "=================== End of collection of collisions "
                    "==================="
                )
            raise StopIteration
        else:
            self._run_at_grid_point()
            self._grid_point_count += 1
            return self._grid_point_count - 1

    @property
    def mesh_numbers(self):
        """Return mesh numbers of GR-grid."""
        return self._pp.mesh_numbers

    def get_mesh_numbers(self):
        """Return mesh numbers of GR-grid."""
        warnings.warn(
            "Use attribute, Conductivity.mesh_numbers "
            "instead of Conductivity.get_mesh_numbers().",
            DeprecationWarning,
        )
        return self.mesh_numbers

    @property
    def bz_grid(self):
        """Return GR-grid."""
        return self._pp.bz_grid

    @property
    def frequencies(self):
        """Return frequencies at grid points.

        Grid points are those at mode kappa are calculated.

        """
        return self._frequencies[self._grid_points]

    def get_frequencies(self):
        """Return frequencies at grid points.

        Grid points are those at mode kappa are calculated.

        """
        warnings.warn(
            "Use attribute, Conductivity.frequencies "
            "instead of Conductivity.get_frequencies().",
            DeprecationWarning,
        )
        return self.frequencies

    @property
    def qpoints(self):
        """Return q-points where mode kappa are calculated."""
        return np.array(
            self._get_qpoint_from_gp_index(self._grid_points),
            dtype="double",
            order="C",
        )

    def get_qpoints(self):
        """Return q-points where mode kappa are calculated."""
        warnings.warn(
            "Use attribute, Conductivity.qpoints "
            "instead of Conductivity.get_qpoints().",
            DeprecationWarning,
        )
        return self.qpoints

    @property
    def grid_points(self):
        """Return grid point indices where mode kappa are calculated.

        Grid point indices are given in BZ-grid.

        """
        return self._grid_points

    def get_grid_points(self):
        """Return grid point indices where mode kappa are calculated.

        Grid point indices are given in BZ-grid.

        """
        warnings.warn(
            "Use attribute, Conductivity.grid_points "
            "instead of Conductivity.get_grid_points().",
            DeprecationWarning,
        )
        return self.grid_points

    @property
    def grid_weights(self):
        """Return grid point weights where mode kappa are calculated."""
        return self._grid_weights

    def get_grid_weights(self):
        """Return grid point weights where mode kappa are calculated."""
        warnings.warn(
            "Use attribute, Conductivity.grid_weights "
            "instead of Conductivity.get_grid_weights().",
            DeprecationWarning,
        )
        return self.grid_weights

    @property
    def temperatures(self):
        """Setter and getter of temperatures."""
        return self._temperatures

    @temperatures.setter
    def temperatures(self, temperatures):
        self._temperatures = np.array(temperatures, dtype="double")
        self._allocate_values()

    def get_temperatures(self):
        """Return temperatures."""
        warnings.warn(
            "Use attribute, Conductivity.temperatures "
            "instead of Conductivity.get_temperatures().",
            DeprecationWarning,
        )
        return self.temperatures

    def set_temperatures(self, temperatures):
        """Set temperatures."""
        warnings.warn(
            "Use attribute, Conductivity.temperatures "
            "instead of Conductivity.set_temperatures().",
            DeprecationWarning,
        )
        self.temperatures = temperatures

    @property
    def gamma(self):
        """Setter and getter of gamma."""
        return self._gamma

    @gamma.setter
    def gamma(self, gamma):
        self._gamma = gamma
        self._read_gamma = True

    def get_gamma(self):
        """Return gamma."""
        warnings.warn(
            "Use attribute, Conductivity.gamma " "instead of Conductivity.get_gamma().",
            DeprecationWarning,
        )
        return self.gamma

    def set_gamma(self, gamma):
        """Set gamma."""
        warnings.warn(
            "Use attribute, Conductivity.gamma " "instead of Conductivity.set_gamma().",
            DeprecationWarning,
        )
        self.gamma = gamma

    @property
    def gamma_isotope(self):
        """Setter and getter of gamma from isotope."""
        return self._gamma_iso

    @gamma_isotope.setter
    def gamma_isotope(self, gamma_iso):
        self._gamma_iso = gamma_iso
        self._read_gamma_iso = True

    def get_gamma_isotope(self):
        """Return gamma from isotope."""
        warnings.warn(
            "Use attribute, Conductivity.gamma_isotope "
            "instead of Conductivity.get_gamma_isotope().",
            DeprecationWarning,
        )
        return self.gamma_isotope

    def set_gamma_isotope(self, gamma_iso):
        """Set gamma from isotope."""
        warnings.warn(
            "Use attribute, Conductivity.gamma_isotope "
            "instead of Conductivity.set_gamma_isotope().",
            DeprecationWarning,
        )
        self.gamma_isotope = gamma_iso

    @property
    def sigmas(self):
        """Return sigmas."""
        return self._sigmas

    def get_sigmas(self):
        """Return sigmas."""
        warnings.warn(
            "Use attribute, Conductivity.sigmas "
            "instead of Conductivity.get_sigmas().",
            DeprecationWarning,
        )
        return self.sigmas

    @property
    def sigma_cutoff_width(self):
        """Return smearing width cutoff."""
        return self._sigma_cutoff

    def get_sigma_cutoff_width(self):
        """Return smearing width cutoff."""
        warnings.warn(
            "Use attribute, Conductivity.sigma_cutoff_width "
            "instead of Conductivity.get_sigma_cutoff_width().",
            DeprecationWarning,
        )
        return self.sigma_cutoff_width

    @property
    def grid_point_count(self):
        """Return interator count of self."""
        return self._grid_point_count

    def get_grid_point_count(self):
        """Return interator count of self."""
        warnings.warn(
            "Use attribute, Conductivity.grid_point_count "
            "instead of Conductivity.get_grid_point_count().",
            DeprecationWarning,
        )
        return self.grid_point_count

    @property
    def averaged_pp_interaction(self):
        """Return averaged pp strength."""
        return self._averaged_pp_interaction

    def get_averaged_pp_interaction(self):
        """Return averaged pp interaction strength."""
        warnings.warn(
            "Use attribute, Conductivity.averaged_pp_interaction "
            "instead of Conductivity.get_averaged_pp_interaction().",
            DeprecationWarning,
        )
        return self.averaged_pp_interaction

    def get_number_of_sampling_grid_points(self):
        """Return number of grid points.

        This is calculated by the sum of numbers of arms of k-start in
        `Conductivity._set_gv_by_gv`.

        """
        return self._num_sampling_grid_points

    def _get_point_operations(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return reciprocal point group operations.

        Returns
        -------
        point_operations : ndarray
            Operations in reduced coordinates.
            shape=(num_operations, 3, 3), dtype='int_'
        rotations_cartesian : ndarray
            Operations in Cartesian coordinates.
            shape=(num_operations, 3, 3), dtype='double'

        """
        if not self._is_kappa_star:
            point_operations = np.array(
                [np.eye(3, dtype="int_")], dtype="int_", order="C"
            )
            rotations_cartesian = np.array(
                [np.eye(3, dtype="double")], dtype="double", order="C"
            )
        else:
            point_operations = self._pp.bz_grid.reciprocal_operations
            rotations_cartesian = self._pp.bz_grid.rotations_cartesian

        return point_operations, rotations_cartesian

    def _get_grid_info(self, grid_points) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Return grid point information in BZGrid.

        Returns
        -------
        grid_points : ndarray
            Grid point indices in BZ-grid to be iterated over.
            shape=(len(grid_points),), dtype='int_'
        ir_grid_points : ndarray
            Irreducible grid points in BZ-grid on regular grid.
            shape=(len(ir_grid_points),), dtype='int_'
        grid_weights : ndarray
            Grid weights of `grid_points`. If grid symmetry is not broken,
            these values are equivalent to numbers of k-star arms.

        """
        ir_grid_points, grid_weights = self._get_ir_grid_points(grid_points)
        if grid_points is not None:  # Specify grid points
            _grid_points = np.array(grid_points, dtype="int_")
            _ir_grid_points = ir_grid_points
            _grid_weights = grid_weights
        elif not self._is_kappa_star:  # All grid points
            _grid_points = self._pp.bz_grid.grg2bzg
            _ir_grid_points = _grid_points
            _grid_weights = np.ones(len(_grid_points), dtype="int_")
        else:  # Automatic sampling
            _grid_points = ir_grid_points
            _ir_grid_points = ir_grid_points
            _grid_weights = grid_weights
        return _grid_points, _ir_grid_points, _grid_weights

    @abstractmethod
    def _run_at_grid_point(self):
        """Run at conductivity calculation at specified grid point.

        Should be implementated in Conductivity* class.

        """
        raise NotImplementedError()

    @abstractmethod
    def _allocate_values(self):
        """Allocate necessary data arrays.

        Should be implementated in Conductivity* class.

        """
        raise NotImplementedError()

    @abstractmethod
    def _set_velocities(self, i_gp, i_data):
        """Set velocities at grid point and at data location.

        Should be implementated in Conductivity*MixIn.

        """
        raise NotImplementedError()

    @abstractmethod
    def _init_velocity(self, gv_delta_q):
        """Initialize velocitiy class instance.

        Should be implementated in Conductivity*MixIn.

        """
        raise NotImplementedError()

    @abstractmethod
    def _set_cv(self, i_gp, i_data):
        """Set heat capacity at grid point and at data location.

        Should be implementated in Conductivity*MixIn.

        """
        raise NotImplementedError()

    def _get_ir_grid_points(self, grid_points):
        """Return ir-grid-points and grid weights in BZGrid."""
        ir_grid_points, ir_grid_weights, ir_grid_map = get_ir_grid_points(
            self._pp.bz_grid
        )
        ir_grid_points = np.array(
            self._pp.bz_grid.grg2bzg[ir_grid_points], dtype="int_"
        )
        if grid_points is None:
            grid_weights = ir_grid_weights
        else:
            weights = np.zeros_like(ir_grid_map)
            for gp in ir_grid_map:
                weights[gp] += 1
            grid_weights = np.array(
                weights[ir_grid_map[self._pp.bz_grid.bzg2grg[grid_points]]],
                dtype="int_",
            )

        return ir_grid_points, grid_weights

    def _get_qpoint_from_gp_index(self, i_gps):
        """Return q-point(s) in reduced coordinates of grid point(s).

        Parameters
        ----------
        i_gps : int or ndarray
            BZ-grid index (int) or indices (ndarray).

        """
        return np.dot(self._pp.bz_grid.addresses[i_gps], self._pp.bz_grid.QDinv.T)

    def _get_multiplicity_at_q(self, i_gp):
        """Return multiplicity (order of site-symmetry) of q-point."""
        if self._is_kappa_star:
            q = self._get_qpoint_from_gp_index(self._grid_points[i_gp])
            reclat = np.linalg.inv(self._pp.primitive.cell)
            multi = 0
            for q_rot in [np.dot(r, q) for r in self._point_operations]:
                diff = q - q_rot
                diff -= np.rint(diff)
                dist = np.linalg.norm(np.dot(reclat, diff))
                if dist < self._pp.primitive_symmetry.tolerance:
                    multi += 1
        else:
            multi = 1
        return multi

    def _get_kstar_order(self, i_gp, multi):
        """Return order (number of arms) of kstar.

        multi : int
            Multiplicity of q-point of `i_gp`, which can be obtained by
            `self._get_multiplicity_at_q(i_gp)`.

        """
        order_kstar = len(self._point_operations) // multi
        if order_kstar != self._grid_weights[i_gp]:
            if self._log_level:
                text = (
                    "Number of elements in k* is unequal "
                    "to number of equivalent grid-points. "
                    "This means that the mesh sampling grids break "
                    "symmetry. Please check carefully "
                    "the convergence over grid point densities."
                )
                msg = textwrap.fill(
                    text, initial_indent=" ", subsequent_indent=" ", width=70
                )
                print("*" * 30 + "Warning" + "*" * 30)
                print(msg)
                print("*" * 67)

        return order_kstar

    def _get_gamma_isotope_at_sigmas(self, i):
        gamma_iso = []
        for sigma in self._sigmas:
            if self._log_level:
                text = "Calculating Gamma of ph-isotope with "
                if sigma is None:
                    text += "tetrahedron method"
                else:
                    text += "sigma=%s" % sigma
                print(text)

            self._isotope.sigma = sigma
            self._isotope.set_phonons(
                self._frequencies,
                self._eigenvectors,
                self._phonon_done,
                dm=self._pp.dynamical_matrix,
            )
            gp = self._grid_points[i]
            self._isotope.set_grid_point(gp)
            self._isotope.run()
            gamma_iso.append(self._isotope.gamma)

        return np.array(gamma_iso, dtype="double", order="C")

    def _set_isotope(self, mass_variances):
        self._isotope = Isotope(
            self._pp.mesh_numbers,
            self._pp.primitive,
            mass_variances=mass_variances,
            bz_grid=self._pp.bz_grid,
            frequency_factor_to_THz=self._pp.frequency_factor_to_THz,
            symprec=self._pp.primitive_symmetry.tolerance,
            cutoff_frequency=self._pp.cutoff_frequency,
            lapack_zheev_uplo=self._pp.lapack_zheev_uplo,
        )
        self._mass_variances = self._isotope.mass_variances

    def _get_main_diagonal(self, i, j, k):
        main_diagonal = self._gamma[j, k, i].copy()
        if self._is_isotope:
            main_diagonal += self._gamma_iso[j, i]
        if self._boundary_mfp is not None:
            main_diagonal += self._get_boundary_scattering(i)
        return main_diagonal

    def _get_boundary_scattering(self, i_gp):
        num_band = len(self._pp.primitive) * 3
        g_boundary = np.zeros(num_band, dtype="double")
        try:
            gv = self._gv
        except AttributeError:
            print("(_get_boundary_scattering) _gv has to be implemented.")
            return g_boundary

        for ll in range(num_band):
            g_boundary[ll] = (
                np.linalg.norm(gv[i_gp, ll])
                * Angstrom
                * 1e6
                / (4 * np.pi * self._boundary_mfp)
            )
        return g_boundary

    def _show_log_header(self, i_gp):
        if self._log_level:
            bzgp = self._grid_points[i_gp]
            print(
                "======================= Grid point %d (%d/%d) "
                "=======================" % (bzgp, i_gp + 1, len(self._grid_points))
            )
            qpoint = self._get_qpoint_from_gp_index(bzgp)
            print("q-point: (%5.2f %5.2f %5.2f)" % tuple(qpoint))
            if self._boundary_mfp is not None:
                if self._boundary_mfp > 1000:
                    print(
                        "Boundary mean free path (millimetre): %.3f"
                        % (self._boundary_mfp / 1000.0)
                    )
                else:
                    print(
                        "Boundary mean free path (micrometre): %.5f"
                        % self._boundary_mfp
                    )
            if self._is_isotope:
                print(
                    (
                        "Mass variance parameters: "
                        + "%5.2e " * len(self._mass_variances)
                    )
                    % tuple(self._mass_variances)
                )
            print(end="", flush=True)



"""Calculation of mode Grueneisen parameters from fc3."""
# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

import numpy as np
from phonopy.harmonic.dynamical_matrix import get_dynamical_matrix
from phonopy.structure.cells import sparse_to_dense_svecs
from phonopy.structure.grid_points import get_qpoints
from phonopy.units import VaspToTHz


def run_gruneisen_parameters(
    fc2,
    fc3,
    supercell,
    primitive,
    band_paths,
    mesh,
    rotations,
    qpoints,
    nac_params=None,
    nac_q_direction=None,
    ion_clamped=False,
    factor=None,
    symprec=1e-5,
    output_filename=None,
    log_level=1,
):
    """Run mode Grueneisen parameter calculation.

    The results is written into files.

    """
    if log_level:
        print("-" * 23 + " Phonon Gruneisen parameter " + "-" * 23)
        if mesh is not None:
            print("Mesh sampling: [ %d %d %d ]" % tuple(mesh))
        elif band_paths is not None:
            print("Paths in reciprocal reduced coordinates:")
            for path in band_paths:
                print(
                    "[%5.2f %5.2f %5.2f] --> [%5.2f %5.2f %5.2f]"
                    % (tuple(path[0]) + tuple(path[-1]))
                )
        if ion_clamped:
            print("To be calculated with ion clamped.")

        sys.stdout.flush()

    gruneisen = Gruneisen(
        fc2,
        fc3,
        supercell,
        primitive,
        nac_params=nac_params,
        nac_q_direction=nac_q_direction,
        ion_clamped=ion_clamped,
        factor=factor,
        symprec=symprec,
    )

    if log_level > 0:
        dm = gruneisen.dynamical_matrix
        if dm.is_nac() and dm.nac_method == "gonze":
            dm.show_Gonze_nac_message()

    if mesh is not None:
        gruneisen.set_sampling_mesh(mesh, rotations=rotations, is_gamma_center=True)
        filename_ext = ".hdf5"
    elif band_paths is not None:
        gruneisen.set_band_structure(band_paths)
        filename_ext = ".yaml"
    elif qpoints is not None:
        gruneisen.set_qpoints(qpoints)
        filename_ext = ".yaml"

    gruneisen.run()

    if output_filename is None:
        filename = "gruneisen"
    else:
        filename = "gruneisen." + output_filename
    gruneisen.write(filename=filename)

    if log_level:
        print("Gruneisen parameters are written in %s" % (filename + filename_ext))


class Gruneisen:
    """Calculat mode Grueneisen parameters from fc3."""

    def __init__(
        self,
        fc2,
        fc3,
        supercell,
        primitive,
        nac_params=None,
        nac_q_direction=None,
        ion_clamped=False,
        factor=VaspToTHz,
        symprec=1e-5,
    ):
        """Init method."""
        self._fc2 = fc2
        self._fc3 = fc3
        self._scell = supercell
        self._pcell = primitive
        self._ion_clamped = ion_clamped
        self._factor = factor
        self._symprec = symprec
        self._dm = get_dynamical_matrix(
            self._fc2,
            self._scell,
            self._pcell,
            nac_params=nac_params,
            symprec=self._symprec,
        )
        self._nac_q_direction = nac_q_direction

        svecs, multi = self._pcell.get_smallest_vectors()
        if self._pcell.store_dense_svecs:
            self._svecs = svecs
            self._multi = multi
        else:
            self._svecs, self._multi = sparse_to_dense_svecs(svecs, multi)

        if self._ion_clamped:
            num_atom_prim = len(self._pcell)
            self._X = np.zeros((num_atom_prim, 3, 3, 3), dtype=float)
        else:
            self._X = self._get_X()
        self._dPhidu = self._get_dPhidu()

        self._gruneisen_parameters = None
        self._frequencies = None
        self._qpoints = None
        self._mesh = None
        self._band_paths = None
        self._band_distances = None
        self._run_mode = None
        self._weights = None

    def run(self):
        """Run mode Grueneisen parameter calculation."""
        if self._run_mode == "band":
            (
                self._gruneisen_parameters,
                self._frequencies,
            ) = self._calculate_band_paths()
        elif self._run_mode == "qpoints" or self._run_mode == "mesh":
            (
                self._gruneisen_parameters,
                self._frequencies,
            ) = self._calculate_at_qpoints(self._qpoints)
        else:
            sys.stderr.write("Q-points are not specified.\n")

    @property
    def dynamical_matrix(self):
        """Return DynamicalMatrix instance."""
        return self._dm

    def get_gruneisen_parameters(self):
        """Return mode Grueneisen paraterms."""
        return self._gruneisen_parameters

    def set_qpoints(self, qpoints):
        """Set q-points."""
        self._run_mode = "qpoints"
        self._qpoints = qpoints

    def set_sampling_mesh(
        self, mesh, rotations=None, shift=None, is_gamma_center=False
    ):
        """Set sampling mesh."""
        self._run_mode = "mesh"
        self._mesh = np.array(mesh, dtype="intc")
        self._qpoints, self._weights = get_qpoints(
            self._mesh,
            np.linalg.inv(self._pcell.cell),
            q_mesh_shift=shift,
            is_gamma_center=is_gamma_center,
            rotations=rotations,
        )

    def set_band_structure(self, paths):
        """Set band structure paths."""
        self._run_mode = "band"
        self._band_paths = paths
        rec_lattice = np.linalg.inv(self._pcell.cell)
        self._band_distances = []
        for path in paths:
            distances_at_path = [0.0]
            for i in range(len(path) - 1):
                distances_at_path.append(
                    np.linalg.norm(np.dot(rec_lattice, path[i + 1] - path[i]))
                    + distances_at_path[-1]
                )
            self._band_distances.append(distances_at_path)

    def write(self, filename="gruneisen"):
        """Write result in a file."""
        if self._gruneisen_parameters is not None:
            if self._run_mode == "band":
                self._write_band_yaml(filename + ".yaml")
            elif self._run_mode == "qpoints":
                self._write_mesh_yaml(filename + ".yaml")
            elif self._run_mode == "mesh":
                self._write_mesh_hdf5(filename + ".hdf5")

    def _write_mesh_yaml(self, filename):
        with open(filename, "w") as f:
            if self._run_mode == "mesh":
                f.write("mesh: [ %5d, %5d, %5d ]\n" % tuple(self._mesh))
            f.write("nqpoint: %d\n" % len(self._qpoints))
            f.write("phonon:\n")
            for i, (q, g_at_q, freqs_at_q) in enumerate(
                zip(self._qpoints, self._gruneisen_parameters, self._frequencies)
            ):
                f.write("- q-position: [ %10.7f, %10.7f, %10.7f ]\n" % tuple(q))
                if self._weights is not None:
                    f.write("  multiplicity: %d\n" % self._weights[i])
                f.write("  band:\n")
                for j, (g, freq) in enumerate(zip(g_at_q, freqs_at_q)):
                    f.write("  - # %d\n" % (j + 1))
                    f.write("    frequency: %15.10f\n" % freq)
                    f.write("    gruneisen: %15.10f\n" % (g.trace() / 3))
                    f.write("    gruneisen_tensor:\n")
                    for g_xyz in g:
                        f.write("    - [ %10.7f, %10.7f, %10.7f ]\n" % tuple(g_xyz))

    def _write_band_yaml(self, filename):
        with open(filename, "w") as f:
            f.write("path:\n\n")
            for path, distances, gs, fs in zip(
                self._band_paths,
                self._band_distances,
                self._gruneisen_parameters,
                self._frequencies,
            ):
                f.write("- nqpoint: %d\n" % len(path))
                f.write("  phonon:\n")
                for i, (q, d, g_at_q, freqs_at_q) in enumerate(
                    zip(path, distances, gs, fs)
                ):  # noqa E125
                    f.write("  - q-position: [ %10.7f, %10.7f, %10.7f ]\n" % tuple(q))
                    f.write("    distance: %10.7f\n" % d)
                    f.write("    band:\n")
                    for j, (g, freq) in enumerate(zip(g_at_q, freqs_at_q)):
                        f.write("    - # %d\n" % (j + 1))
                        f.write("      frequency: %15.10f\n" % freq)
                        f.write("      gruneisen: %15.10f\n" % (g.trace() / 3))
                        f.write("      gruneisen_tensor:\n")
                        for g_xyz in g:
                            f.write(
                                "      - [ %10.7f, %10.7f, %10.7f ]\n" % tuple(g_xyz)
                            )
                    f.write("\n")

    def _write_mesh_hdf5(self, filename="gruneisen.hdf5"):
        import h5py

        g = self._gruneisen_parameters
        gruneisen = np.array(
            (g[:, :, 0, 0] + g[:, :, 1, 1] + g[:, :, 2, 2]) / 3,
            dtype="double",
            order="C",
        )

        with h5py.File(filename, "w") as w:
            w.create_dataset("mesh", data=self._mesh)
            w.create_dataset("gruneisen", data=gruneisen)
            w.create_dataset("gruneisen_tensor", data=self._gruneisen_parameters)
            w.create_dataset("weight", data=self._weights)
            w.create_dataset("frequency", data=self._frequencies)
            w.create_dataset("qpoint", data=self._qpoints)

    def _calculate_at_qpoints(self, qpoints):
        gruneisen_parameters = []
        frequencies = []
        for i, q in enumerate(qpoints):
            if self._dm.is_nac():
                if (np.abs(q) < 1e-5).all():  # If q is almost at Gamma
                    if self._run_mode == "band":
                        # Direction estimated from neighboring point
                        if i > 0:
                            q_dir = qpoints[i] - qpoints[i - 1]
                        elif i == 0 and len(qpoints) > 1:
                            q_dir = qpoints[i + 1] - qpoints[i]
                        else:
                            q_dir = None
                        g, omega2 = self._get_gruneisen_tensor(q, nac_q_direction=q_dir)
                    else:  # Specified q-vector
                        g, omega2 = self._get_gruneisen_tensor(
                            q, nac_q_direction=self._nac_q_direction
                        )
                else:  # If q is away from Gamma-point, then q-vector
                    g, omega2 = self._get_gruneisen_tensor(q, nac_q_direction=q)
            else:  # Without NAC
                g, omega2 = self._get_gruneisen_tensor(q)
            gruneisen_parameters.append(g)
            frequencies.append(np.sqrt(abs(omega2)) * np.sign(omega2) * self._factor)

        return (
            np.array(gruneisen_parameters, dtype="double", order="C"),
            np.array(frequencies, dtype="double", order="C"),
        )

    def _calculate_band_paths(self):
        gruneisen_parameters = []
        frequencies = []
        for path in self._band_paths:
            (gruneisen_at_path, frequencies_at_path) = self._calculate_at_qpoints(path)
            gruneisen_parameters.append(gruneisen_at_path)
            frequencies.append(frequencies_at_path)

        return gruneisen_parameters, frequencies

    def _get_gruneisen_tensor(self, q, nac_q_direction=None):
        if nac_q_direction is None:
            self._dm.run(q)
        else:
            self._dm.run(q, nac_q_direction)
        omega2, w = np.linalg.eigh(self._dm.dynamical_matrix)
        g = np.zeros((len(omega2), 3, 3), dtype=float)
        num_atom_prim = len(self._pcell)
        dDdu = self._get_dDdu(q)

        for s in range(len(omega2)):
            if (np.abs(q) < 1e-5).all() and s < 3:
                continue
            for i in range(3):
                for j in range(3):
                    for nu in range(num_atom_prim):
                        for pi in range(num_atom_prim):
                            g[s] += (
                                w[nu * 3 + i, s].conjugate()
                                * dDdu[nu, pi, i, j]
                                * w[pi * 3 + j, s]
                            ).real

            g[s] *= -1.0 / 2 / omega2[s]

        return g, omega2

    def _get_dDdu(self, q):
        num_atom_prim = len(self._pcell)
        p2s = self._pcell.p2s_map
        s2p = self._pcell.s2p_map
        m = self._pcell.masses
        dPhidu = self._dPhidu
        dDdu = np.zeros((num_atom_prim, num_atom_prim, 3, 3, 3, 3), dtype=complex)

        for nu in range(num_atom_prim):
            for pi, p in enumerate(p2s):
                for Ppi, s in enumerate(s2p):
                    if not s == p:
                        continue
                    adrs = self._multi[Ppi, nu][1]
                    multi = self._multi[Ppi, nu][0]
                    phase = (
                        np.exp(
                            2j
                            * np.pi
                            * np.dot(self._svecs[adrs : (adrs + multi), :], q)
                        ).sum()
                        / multi
                    )
                    dDdu[nu, pi] += phase * dPhidu[nu, Ppi]
                dDdu[nu, pi] /= np.sqrt(m[nu] * m[pi])

        return dDdu

    def _get_dPhidu(self):
        fc3 = self._fc3
        num_atom_prim = len(self._pcell)
        num_atom_super = len(self._scell)
        p2s = self._pcell.p2s_map
        dPhidu = np.zeros((num_atom_prim, num_atom_super, 3, 3, 3, 3), dtype=float)

        for nu in range(num_atom_prim):
            Y = self._get_Y(nu)
            for pi in range(num_atom_super):
                for i in range(3):
                    for j in range(3):
                        for k in range(3):
                            for ll in range(3):
                                for m in range(3):
                                    dPhidu[nu, pi, i, j, k, ll] = (
                                        fc3[p2s[nu], pi, :, i, j, :] * Y[:, :, k, ll]
                                    ).sum()
                                    # Symmetrization?
                                    # (Y[:,:,k,l] + Y[:,:,l,k]) / 2).sum()

        return dPhidu

    def _get_Y(self, nu):
        X = self._X
        lat = self._pcell.cell
        num_atom_super = len(self._scell)
        R = self._get_R(num_atom_super, nu, lat)
        s2p = self._pcell.s2p_map
        p2p = self._pcell.p2p_map

        Y = np.zeros((num_atom_super, 3, 3, 3), dtype=float)

        for Mmu in range(num_atom_super):
            for i in range(3):
                Y[Mmu, i, i, :] = R[Mmu, :]
            Y[Mmu] += X[p2p[s2p[Mmu]]]

        return Y

    def _get_X(self):
        num_atom_super = len(self._scell)
        num_atom_prim = len(self._pcell)
        p2s = self._pcell.p2s_map
        lat = self._pcell.cell
        X = np.zeros((num_atom_prim, 3, 3, 3), dtype=float)
        G = self._get_Gamma()
        P = self._fc2

        for mu in range(num_atom_prim):
            for nu in range(num_atom_prim):
                R = self._get_R(num_atom_super, nu, lat)
                for i in range(3):
                    for j in range(3):
                        for k in range(3):
                            for ll in range(3):
                                X[mu, i, j, k] -= G[mu, nu, i, ll] * np.dot(
                                    P[p2s[nu], :, ll, j], R[:, k]
                                )

        return X

    def _get_R(self, num_atom_super, nu, lat):
        R = []
        for Npi in range(num_atom_super):
            adrs = self._multi[Npi, nu][1]
            multi = self._multi[Npi, nu][0]
            R.append(
                np.dot(self._svecs[adrs : (adrs + multi), :].sum(axis=0) / multi, lat)
            )
        return np.array(R)

    def _get_Gamma(self):
        num_atom_prim = len(self._pcell)
        m = self._pcell.masses
        self._dm.run([0, 0, 0])
        vals, vecs = np.linalg.eigh(self._dm.dynamical_matrix.real)
        G = np.zeros((num_atom_prim, num_atom_prim, 3, 3), dtype=float)

        for pi in range(num_atom_prim):
            for mu in range(num_atom_prim):
                for k in range(3):
                    for i in range(3):
                        # Eigenvectors are real.
                        # 3: means optical modes
                        G[pi, mu, k, i] = (
                            1.0
                            / np.sqrt(m[pi] * m[mu])
                            * (
                                vecs[pi * 3 + k, 3:] * vecs[mu * 3 + i, 3:] / vals[3:]
                            ).sum()
                        )
        return G



"""API for isotope scattering."""
# Copyright (C) 2019 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np
from phonopy.units import VaspToTHz

from phono3py.other.isotope import Isotope


class Phono3pyIsotope:
    """Class to calculate isotope scattering."""

    def __init__(
        self,
        mesh,
        primitive,
        mass_variances=None,  # length of list is num_atom.
        band_indices=None,
        sigmas=None,
        frequency_factor_to_THz=VaspToTHz,
        use_grg=False,
        store_dense_gp_map=True,
        symprec=1e-5,
        cutoff_frequency=None,
        lapack_zheev_uplo="L",
    ):
        """Init method."""
        if sigmas is None:
            self._sigmas = [
                None,
            ]
        else:
            self._sigmas = sigmas

        if not store_dense_gp_map:
            warnings.warn(
                "Phono3pyIsotope init parameter of store_dense_gp_map is deprecated. "
                "This will be always set True.",
                DeprecationWarning,
            )

        self._iso = Isotope(
            mesh,
            primitive,
            mass_variances=mass_variances,
            band_indices=band_indices,
            frequency_factor_to_THz=frequency_factor_to_THz,
            use_grg=use_grg,
            store_dense_gp_map=store_dense_gp_map,
            symprec=symprec,
            cutoff_frequency=cutoff_frequency,
            lapack_zheev_uplo=lapack_zheev_uplo,
        )

    @property
    def dynamical_matrix(self):
        """Return dynamical matrix class instance."""
        return self._iso.dynamical_matrix

    @property
    def grid(self):
        """Return BZGrid class instance."""
        return self._iso.bz_grid

    @property
    def gamma(self):
        """Return calculated isotope scattering."""
        return self._gamma

    @property
    def frequencies(self):
        """Return phonon frequencies at grid points."""
        return self._iso.get_phonons()[0][self._grid_points]

    @property
    def grid_points(self):
        """Return grid points in BZ-grid."""
        return self._grid_points

    def run(self, grid_points, lang="C"):
        """Calculate isotope scattering."""
        gamma = np.zeros(
            (len(self._sigmas), len(grid_points), len(self._iso.band_indices)),
            dtype="double",
        )
        self._grid_points = np.array(grid_points, dtype="int_")

        for j, gp in enumerate(grid_points):
            self._iso.set_grid_point(gp)

            print("--------------- Isotope scattering ---------------")
            print("Grid point: %d" % gp)
            adrs = self._iso.bz_grid.addresses[gp]
            bz_grid = self._iso.bz_grid
            print(bz_grid.D_diag)
            q = np.dot(adrs.astype("double") / bz_grid.D_diag, bz_grid.Q.T)
            print("q-point: %s" % q)

            if self._sigmas:
                for i, sigma in enumerate(self._sigmas):
                    if sigma is None:
                        print("Tetrahedron method")
                    else:
                        print("Sigma: %s" % sigma)
                    self._iso.sigma = sigma
                    self._iso.run(lang=lang)
                    gamma[i, j] = self._iso.gamma
                    frequencies = self._iso.get_phonons()[0]
                    print("")
                    print("Phonon-isotope scattering rate in THz (1/4pi-tau)")
                    print(" Frequency     Rate")
                    for g, f in zip(self._iso.gamma, frequencies[gp]):
                        print("%8.3f     %5.3e" % (f, g))
            else:
                print("sigma or tetrahedron method has to be set.")
        self._gamma = gamma

    def init_dynamical_matrix(
        self,
        fc2,
        supercell,
        primitive,
        nac_params=None,
        frequency_scale_factor=None,
        decimals=None,
    ):
        """Initialize dynamical matrix."""
        self._primitive = primitive
        self._iso.init_dynamical_matrix(
            fc2,
            supercell,
            primitive,
            nac_params=nac_params,
            frequency_scale_factor=frequency_scale_factor,
            decimals=decimals,
        )

    def set_sigma(self, sigma):
        """Set sigma. None means tetrahedron method."""
        self._iso.set_sigma(sigma)



"""Launch script of ZnTe AiiDA calculation using aiida-phononpy."""
from aiida.engine import submit
from aiida.manage.configuration import load_profile
from aiida.orm import Bool, Float, Str
from aiida.plugins import DataFactory, WorkflowFactory
from aiida_phonopy.common.utils import phonopy_atoms_to_structure
from phonopy.interface.vasp import read_vasp_from_strings

load_profile()

Dict = DataFactory("dict")


def get_settings(cutoff_energy, is_nac=False):
    """Set up parameters."""
    unitcell_str = """  Zn Te
   1.0
     6.0653118499999996    0.0000000000000000    0.0000000000000000
     0.0000000000000000    6.0653118499999996    0.0000000000000000
     0.0000000000000000    0.0000000000000000    6.0653118499999996
 Zn Te
   4   4
Direct
   0.0000000000000000  0.0000000000000000  0.0000000000000000
   0.0000000000000000  0.5000000000000000  0.5000000000000000
   0.5000000000000000  0.0000000000000000  0.5000000000000000
   0.5000000000000000  0.5000000000000000  0.0000000000000000
   0.2500000000000000  0.2500000000000000  0.7500000000000000
   0.2500000000000000  0.7500000000000000  0.2500000000000000
   0.7500000000000000  0.2500000000000000  0.2500000000000000
   0.7500000000000000  0.7500000000000000  0.7500000000000000"""

    cell = read_vasp_from_strings(unitcell_str)
    structure = phonopy_atoms_to_structure(cell)

    base_incar_dict = {
        "PREC": "Accurate",
        "IBRION": -1,
        "EDIFF": 1e-8,
        "NELMIN": 5,
        "NELM": 100,
        "ENCUT": cutoff_energy,
        "IALGO": 38,
        "ISMEAR": 0,
        "SIGMA": 0.01,
        "GGA": "PS",
        "LREAL": False,
        "lcharg": False,
        "lwave": False,
    }

    code_string = "vasp544mpi@nancy"
    resources = {"parallel_env": "mpi*", "tot_num_mpiprocs": 24}

    base_config = {
        "code_string": code_string,
        "potential_family": "PBE.54",
        "potential_mapping": {"Zn": "Zn", "Te": "Te"},
        "options": {"resources": resources, "max_wallclock_seconds": 3600 * 10},
    }
    base_parser_settings = {
        "add_energies": True,
        "add_forces": True,
        "add_stress": True,
    }
    forces_config = base_config.copy()
    kpoints_mesh = [2, 2, 2]
    forces_config.update(
        {
            "kpoints_mesh": kpoints_mesh,
            "kpoints_offset": [0.5, 0.5, 0.5],
            "parser_settings": base_parser_settings,
            "parameters": {"incar": base_incar_dict.copy()},
        }
    )
    forces_config["parameters"]["incar"]["NPAR"] = 4
    nac_config = {
        "code_string": code_string,
        "potential_family": "PBE.54",
        "potential_mapping": {"Zn": "Zn", "Te": "Te"},
        "options": {"resources": resources, "max_wallclock_seconds": 3600 * 10},
    }
    nac_parser_settings = {"add_born_charges": True, "add_dielectrics": True}
    nac_parser_settings.update(base_parser_settings)
    nac_incar_dict = {"lepsilon": True}
    nac_incar_dict.update(base_incar_dict.copy())
    nac_config.update(
        {
            "kpoints_mesh": [8, 8, 8],
            "kpoints_offset": [0.5, 0.5, 0.5],
            "parser_settings": nac_parser_settings,
            "parameters": {"incar": nac_incar_dict},
        }
    )
    phonon_settings = {"supercell_matrix": [2, 2, 2], "distance": 0.03}
    if is_nac:
        phonon_settings["is_nac"] = is_nac

    return structure, forces_config, nac_config, phonon_settings


def launch_phono3py(cutoff_energy=350, is_nac=False):
    """Launch calculation."""
    structure, forces_config, nac_config, phonon_settings = get_settings(
        cutoff_energy, is_nac
    )
    Phono3pyWorkChain = WorkflowFactory("phonopy.phono3py")
    builder = Phono3pyWorkChain.get_builder()
    builder.structure = structure
    builder.calculator_settings = Dict(
        dict={"forces": forces_config, "nac": nac_config}
    )
    builder.run_phono3py = Bool(False)
    builder.remote_phono3py = Bool(False)
    builder.code_string = Str("phonopy@nancy")
    builder.phonon_settings = Dict(dict=phonon_settings)
    builder.symmetry_tolerance = Float(1e-5)
    builder.options = Dict(dict=forces_config["options"])
    dim = phonon_settings["supercell_matrix"]
    kpoints_mesh = forces_config["kpoints_mesh"]
    label = "ZnTe phono3py %dx%dx%d kpt %dx%dx%d PBEsol %d eV" % (
        tuple(dim) + tuple(kpoints_mesh) + (cutoff_energy,)
    )
    builder.metadata.label = label
    builder.metadata.description = label

    future = submit(builder)
    print(label)
    print(future)
    print("Running workchain with pk={}".format(future.pk))


if __name__ == "__main__":
    launch_phono3py(cutoff_energy=500, is_nac=True)



"""Phono3py loader."""
# Copyright (C) 2019 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phono3py project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import os
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
import phonopy.cui.load_helper as load_helper
from phonopy.harmonic.force_constants import show_drift_force_constants
from phonopy.interface.calculator import get_default_physical_units
from phonopy.structure.cells import determinant

from phono3py import Phono3py
from phono3py.cui.create_force_constants import forces_in_dataset, parse_forces
from phono3py.file_IO import read_fc2_from_hdf5, read_fc3_from_hdf5
from phono3py.interface.phono3py_yaml import Phono3pyYaml
from phono3py.phonon3.fc3 import show_drift_fc3


def load(
    phono3py_yaml=None,  # phono3py.yaml-like must be the first argument.
    supercell_matrix=None,
    primitive_matrix=None,
    phonon_supercell_matrix=None,
    is_nac=True,
    calculator=None,
    unitcell=None,
    supercell=None,
    nac_params=None,
    unitcell_filename=None,
    supercell_filename=None,
    born_filename=None,
    forces_fc3_filename: Optional[Union[os.PathLike, Sequence]] = None,
    forces_fc2_filename: Optional[Union[os.PathLike, Sequence]] = None,
    fc3_filename=None,
    fc2_filename=None,
    fc_calculator=None,
    fc_calculator_options=None,
    factor=None,
    produce_fc=True,
    is_symmetry=True,
    symmetrize_fc=True,
    is_mesh_symmetry=True,
    is_compact_fc=False,
    use_grg=False,
    store_dense_gp_map=True,
    store_dense_svecs=True,
    symprec=1e-5,
    log_level=0,
) -> Phono3py:
    """Create Phono3py instance from parameters and/or input files.

    "phono3py_yaml"-like file is parsed unless crystal structure information
    is given by unitcell_filename, supercell_filename, unitcell
    (PhonopyAtoms-like), or supercell (PhonopyAtoms-like).
    Even when "phono3py_yaml"-like file is parse, parameters except for
    crystal structure can be overwritten.

    'fc3.hdf5' is read if found in current directory.
    Unless 'fc3.hdf5' is found and if 'FORCES_FC3' and 'disp_fc3.yaml" are
    found, these are read and fc3 and fc2 are produced.

    if 'fc2.hdf5' is found, this is read.
    Unless 'fc2.hdf5' is found and if 'FORCES_FC2' and 'disp_fc2.yaml" are
    found, these are read and fc2 is produced.

    When force_sets_filename and force_constants_filename are not given,
    'FORCES_FC3' and 'FORCES_FC2' are looked for in the current directory
    as the default behaviour. When 'FORCES_FC3' ('FORCES_FC2') is given in
    the type-1 format, 'disp_fc3.yaml' ('disp_fc2.yaml') is also necessary
    and read.

    Crystal structure
    -----------------
    Means to provide crystal structure(s) and their priority:
        1. unitcell_filename (with supercell_matrix)
        2. supercell_filename
        3. unitcell (with supercell_matrix)
        4. supercell.
        5. phono3py_yaml-like

    Force sets or force constants
    -----------------------------
    Optional. Means to provide information to generate force constants
    and their priority:
        1. fc3_filename (fc2_filename)
        2. forces_fc3_filename (forces_fc2_filename). Do not forget that
           for type-1 format, disp_fc3.yaml (disp_fc2.yaml) has to be given,
           too.
        3. 'fc3.hdf5' and 'fc2.hdf5' are searched in current directory.
        4. 'FORCES_FC3' and 'FORCES_FC2' are searched in current directory.
           'FORCES_FC2' is optional. For type-1 format, 'disp_fc3.yaml' and
           optionally 'disp_fc2.yaml' are also searched in current
           directory. When 'FORCES_FC2' is not found, 'FORCES_FC3' is used
           to create fc2.

    Parameters for non-analytical term correctiion (NAC)
    ----------------------------------------------------
    Optional. Means to provide NAC parameters and their priority:
        1. born_filename
        2. nac_params
        3. phono3py_yaml_like.nac_params if existed and is_nac=True.
        4. 'BORN' is searched in current directory when is_nac=True.

    Parameters
    ----------
    phono3py_yaml : str, optional
        Filename of "phono3py.yaml"-like file. If this is given, the data
        in the file are parsed. Default is None.
    supercell_matrix : array_like, optional
        Supercell matrix multiplied to input cell basis vectors.
        shape=(3, ) or (3, 3), where the former is considered a diagonal
        matrix. Default is the unit matrix.
        dtype=int
    primitive_matrix : array_like or str, optional
        Primitive matrix multiplied to input cell basis vectors. Default is
        the identity matrix.
        When given as array_like, shape=(3, 3), dtype=float.
        When 'F', 'I', 'A', 'C', or 'R' is given instead of a 3x3 matrix,
        the primitive matrix defined at
        https://spglib.github.io/spglib/definition.html
        is used.
        When 'auto' is given, the centring type ('F', 'I', 'A', 'C', 'R', or
        primitive 'P') is automatically chosen.
        Default is 'auto'.
    phonon_supercell_matrix : array_like, optional
        Supercell matrix used for fc2. In phono3py, supercell matrix for fc3
        and fc2 can be different to support longer range interaction of fc2
        than that of fc3. Unless setting this, supercell_matrix is used.
        This is only valide when unitcell or unitcell_filename is given.
        Default is None.
    is_nac : bool, optional
        If True, look for 'BORN' file. If False, NAS is turned off.
        Default is True.
    calculator : str, optional.
        Calculator used for computing forces. This is used to switch the set
        of physical units. Default is None, which is equivalent to "vasp".
    unitcell : PhonopyAtoms, optional
        Input unit cell. Default is None.
    supercell : PhonopyAtoms, optional
        Input supercell. With given, default value of primitive_matrix is set
        to 'auto' (can be overwitten). supercell_matrix is ignored. Default is
        None.
    nac_params : dict, optional
        Parameters required for non-analytical term correction. Default is
        None.
        {'born': Born effective charges
                 (array_like, shape=(primitive cell atoms, 3, 3), dtype=float),
         'dielectric': Dielectric constant matrix
                       (array_like, shape=(3, 3), dtype=float),
         'factor': unit conversion facotr (float)}
    unitcell_filename : os.PathLike, optional
        Input unit cell filename. Default is None.
    supercell_filename : os.PathLike, optional
        Input supercell filename. When this is specified, supercell_matrix is
        ignored. Default is None.
    born_filename : os.PathLike, optional
        Filename corresponding to 'BORN', a file contains non-analytical term
        correction parameters.
    forces_fc3_filename : sequence or os.PathLike, optional
        A two-elemental sequence of filenames corresponding to
        ('FORCES_FC3', 'disp_fc3.yaml') in the type-1 format or a filename
        (os.PathLike) corresponding to 'FORCES_FC3' in the type-2 format.
        Default is None.
    forces_fc2_filename : os.PathLike or sequence, optional
        A two-elemental sequence of filenames corresponding to
        ('FORCES_FC2', 'disp_fc2.yaml') in the type-1 format or a filename
        (os.PathLike) corresponding to 'FORCES_FC2' in the type-2 format.
        Default is None.
    fc3_filename : os.PathLike, optional
        Filename of a file corresponding to 'fc3.hdf5', a file contains
        third-order force constants. Default is None.
    fc2_filename : os.PathLike, optional
        Filename of a file corresponding to 'fc2.hdf5', a file contains
        second-order force constants. Default is None.
    fc_calculator : str, optional
        Force constants calculator. Currently only 'alm'. Default is None.
    fc_calculator_options : str, optional
        Optional parameters that are passed to the external fc-calculator.
        This is given as one text string. How to parse this depends on the
        fc-calculator. For alm, each parameter is splitted by comma ',',
        and each set of key and value pair is written in 'key = value'.
    factor : float, optional
        Phonon frequency unit conversion factor. Unless specified, default
        unit conversion factor for each calculator is used.
    produce_fc : bool, optional
        Setting False, force constants are not calculated from displacements
        and forces. Default is True.
    is_symmetry : bool, optional
        Setting False, crystal symmetry except for lattice translation is not
        considered. Default is True.
    symmetrize_fc : bool, optional
        Setting False, force constants are not symmetrized when creating
        force constants from displacements and forces. Default is True.
    is_mesh_symmetry : bool, optional
        Setting False, reciprocal mesh symmetry is not considered.
        Default is True.
    is_compact_fc : bool, optional
        fc3 are created in the array whose shape is
            True: (primitive, supercell, supecell, 3, 3, 3)
            False: (supercell, supercell, supecell, 3, 3, 3)
        and for fc2
            True: (primitive, supecell, 3, 3)
            False: (supercell, supecell, 3, 3)
        where 'supercell' and 'primitive' indicate number of atoms in these
        cells. Default is False.
    use_grg : bool, optional
        Use generalized regular grid when True. Default is False.
    store_dense_gp_map : bool, optional, Deprecated
        Use new format of BZ grid system. Default is True.
    store_dense_svecs : bool, optional, Deprecated
        Shortest vectors are stored in the dense array format. This is
        expected to be always True. Setting False is for rough
        compatibility with v1.x. Default is True.
    symprec : float, optional
        Tolerance used to find crystal symmetry. Default is 1e-5.
    log_level : int, optional
        Verbosity control. Default is 0.

    """
    if (
        supercell is not None
        or supercell_filename is not None
        or unitcell is not None
        or unitcell_filename is not None
    ):  # noqa E129
        cell, smat, pmat = load_helper.get_cell_settings(
            supercell_matrix=supercell_matrix,
            primitive_matrix=primitive_matrix,
            unitcell=unitcell,
            supercell=supercell,
            unitcell_filename=unitcell_filename,
            supercell_filename=supercell_filename,
            calculator=calculator,
            symprec=symprec,
        )
        if phonon_supercell_matrix is not None:
            if unitcell is None and unitcell_filename is None:
                msg = (
                    "phonon_supercell_matrix can be used only when "
                    "unitcell or unitcell_filename is given."
                )
                raise RuntimeError(msg)
            ph_smat = phonon_supercell_matrix
        else:
            ph_smat = None
        _nac_params = nac_params
        ph3py_yaml = None
    elif phono3py_yaml is not None:
        ph3py_yaml = Phono3pyYaml()
        ph3py_yaml.read(phono3py_yaml)
        cell = ph3py_yaml.unitcell
        smat = ph3py_yaml.supercell_matrix
        ph_smat = ph3py_yaml.phonon_supercell_matrix
        if smat is None:
            smat = np.eye(3, dtype="intc", order="C")
        if primitive_matrix == "auto":
            pmat = "auto"
        else:
            pmat = ph3py_yaml.primitive_matrix

        if nac_params is not None:
            _nac_params = nac_params
        elif is_nac:
            _nac_params = ph3py_yaml.nac_params
        else:
            _nac_params = None

    # units keywords: factor, nac_factor, distance_to_A
    physical_units = get_default_physical_units(calculator)
    if factor is None:
        _factor = physical_units["factor"]
    else:
        _factor = factor
    ph3py = Phono3py(
        cell,
        smat,
        primitive_matrix=pmat,
        phonon_supercell_matrix=ph_smat,
        frequency_factor_to_THz=_factor,
        symprec=symprec,
        is_symmetry=is_symmetry,
        is_mesh_symmetry=is_mesh_symmetry,
        use_grg=use_grg,
        store_dense_gp_map=store_dense_gp_map,
        store_dense_svecs=store_dense_svecs,
        calculator=calculator,
        log_level=log_level,
    )

    # NAC params
    if born_filename is not None or _nac_params is not None or is_nac:
        ph3py.nac_params = load_helper.get_nac_params(
            ph3py.primitive,
            _nac_params,
            born_filename,
            is_nac,
            physical_units["nac_factor"],
            log_level=log_level,
        )

    set_dataset_and_force_constants(
        ph3py,
        ph3py_yaml,
        fc3_filename=fc3_filename,
        fc2_filename=fc2_filename,
        forces_fc3_filename=forces_fc3_filename,
        forces_fc2_filename=forces_fc2_filename,
        phono3py_yaml_filename=phono3py_yaml,
        log_level=log_level,
    )

    if produce_fc:
        compute_force_constants_from_datasets(
            ph3py=ph3py,
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
            symmetrize_fc=symmetrize_fc,
            is_compact_fc=is_compact_fc,
            log_level=log_level,
        )

    if log_level and ph3py.fc3 is not None:
        show_drift_fc3(ph3py.fc3, primitive=ph3py.primitive)
    if log_level and ph3py.fc2 is not None:
        show_drift_force_constants(
            ph3py.fc2, primitive=ph3py.phonon_primitive, name="fc2"
        )

    return ph3py


def set_dataset_and_force_constants(
    ph3py: Phono3py,
    ph3py_yaml: Optional[Phono3pyYaml] = None,
    fc3_filename: Optional[os.PathLike] = None,
    fc2_filename: Optional[os.PathLike] = None,
    forces_fc3_filename: Optional[Union[os.PathLike, Sequence]] = None,
    forces_fc2_filename: Optional[Union[os.PathLike, Sequence]] = None,
    phono3py_yaml_filename: Optional[os.PathLike] = None,
    cutoff_pair_distance: Optional[float] = None,
    log_level: int = 0,
):
    """Set displacements, forces, and create force constants."""
    read_fc = {"fc2": False, "fc3": False}
    read_fc["fc3"] = _set_dataset_or_fc3(
        ph3py,
        ph3py_yaml=ph3py_yaml,
        fc3_filename=fc3_filename,
        forces_fc3_filename=forces_fc3_filename,
        phono3py_yaml_filename=phono3py_yaml_filename,
        cutoff_pair_distance=cutoff_pair_distance,
        log_level=log_level,
    )
    read_fc["fc2"] = _set_dataset_phonon_dataset_or_fc2(
        ph3py,
        ph3py_yaml=ph3py_yaml,
        fc2_filename=fc2_filename,
        forces_fc2_filename=forces_fc2_filename,
        log_level=log_level,
    )

    # Cases that dataset is in phono3py.yaml but not forces.
    if ph3py.dataset is None:
        if ph3py_yaml is not None and ph3py_yaml.dataset is not None:
            ph3py.dataset = ph3py_yaml.dataset
        if ph3py_yaml is not None and ph3py_yaml.phonon_dataset is not None:
            ph3py.phonon_dataset = ph3py_yaml.phonon_dataset

    return read_fc


def compute_force_constants_from_datasets(
    ph3py: Phono3py,
    fc_calculator: Optional[str] = None,
    fc_calculator_options: Optional[dict] = None,
    symmetrize_fc: bool = True,
    is_compact_fc: bool = True,
    log_level: int = 0,
):
    """Compute force constants from datasets."""
    if ph3py.dataset:
        ph3py.produce_fc3(
            symmetrize_fc3r=symmetrize_fc,
            is_compact_fc=is_compact_fc,
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
        )
        if log_level and symmetrize_fc:
            print("fc3 was symmetrized.")

    if ph3py.dataset or ph3py.phonon_dataset:
        if (
            ph3py.phonon_supercell_matrix is None
            and fc_calculator == "alm"
            and ph3py.fc2 is not None
        ):
            if log_level:
                print("fc2 that was fit simultaneously with fc3 by ALM is used.")
        else:
            ph3py.produce_fc2(
                symmetrize_fc2=symmetrize_fc,
                is_compact_fc=is_compact_fc,
                fc_calculator=fc_calculator,
                fc_calculator_options=fc_calculator_options,
            )
            if log_level and symmetrize_fc:
                print("fc2 was symmetrized.")


def _set_dataset_or_fc3(
    ph3py: Phono3py,
    ph3py_yaml: Optional[Phono3pyYaml] = None,
    fc3_filename: Optional[os.PathLike] = None,
    forces_fc3_filename: Optional[Union[os.PathLike, Sequence]] = None,
    phono3py_yaml_filename: Optional[os.PathLike] = None,
    cutoff_pair_distance: Optional[float] = None,
    log_level: int = 0,
) -> bool:
    p2s_map = ph3py.primitive.p2s_map
    read_fc3 = False
    if fc3_filename is not None:
        fc3 = read_fc3_from_hdf5(filename=fc3_filename, p2s_map=p2s_map)
        _check_fc3_shape(ph3py, fc3, filename=fc3_filename)
        ph3py.fc3 = fc3
        read_fc3 = True
        if log_level:
            print('fc3 was read from "%s".' % fc3_filename)
    elif forces_fc3_filename is not None:
        if isinstance(forces_fc3_filename, os.PathLike):
            force_filename = forces_fc3_filename
            disp_filename = None
        else:
            force_filename, disp_filename = forces_fc3_filename
        _set_dataset_for_fc3(
            ph3py,
            ph3py_yaml,
            force_filename,
            disp_filename,
            phono3py_yaml_filename,
            cutoff_pair_distance,
            log_level,
        )
    elif os.path.isfile("fc3.hdf5"):
        fc3 = read_fc3_from_hdf5(filename="fc3.hdf5", p2s_map=p2s_map)
        _check_fc3_shape(ph3py, fc3)
        ph3py.fc3 = fc3
        read_fc3 = True
        if log_level:
            print('fc3 was read from "fc3.hdf5".')
    elif os.path.isfile("FORCES_FC3"):
        disp_filename = None
        if os.path.isfile("disp_fc3.yaml"):
            if ph3py_yaml is None:
                disp_filename = "disp_fc3.yaml"
            elif ph3py_yaml.dataset is None:
                disp_filename = "disp_fc3.yaml"
        _set_dataset_for_fc3(
            ph3py,
            ph3py_yaml,
            "FORCES_FC3",
            disp_filename,
            phono3py_yaml_filename,
            cutoff_pair_distance,
            log_level,
        )
    elif (
        ph3py_yaml is not None
        and ph3py_yaml.dataset is not None
        and forces_in_dataset(ph3py_yaml.dataset)
    ):
        _set_dataset_for_fc3(
            ph3py,
            ph3py_yaml,
            None,
            None,
            phono3py_yaml_filename,
            cutoff_pair_distance,
            log_level,
        )
    return read_fc3


def _set_dataset_phonon_dataset_or_fc2(
    ph3py: Phono3py,
    ph3py_yaml: Optional[Phono3pyYaml] = None,
    fc2_filename: Optional[os.PathLike] = None,
    forces_fc2_filename: Optional[Union[os.PathLike, Sequence]] = None,
    log_level: int = 0,
) -> bool:
    phonon_p2s_map = ph3py.phonon_primitive.p2s_map
    read_fc2 = False
    if fc2_filename is not None:
        fc2 = read_fc2_from_hdf5(filename=fc2_filename, p2s_map=phonon_p2s_map)
        _check_fc2_shape(ph3py, fc2, filename=fc2_filename)
        ph3py.fc2 = fc2
        read_fc2 = True
        if log_level:
            print('fc2 was read from "%s".' % fc2_filename)
    elif forces_fc2_filename is not None:
        if isinstance(forces_fc2_filename, os.PathLike):
            force_filename = forces_fc2_filename
            disp_filename = None
        else:
            force_filename, disp_filename = forces_fc2_filename
        _set_dataset_for_fc2(
            ph3py,
            ph3py_yaml,
            force_filename,
            disp_filename,
            "phonon_fc2",
            log_level,
        )
    elif os.path.isfile("fc2.hdf5"):
        fc2 = read_fc2_from_hdf5(filename="fc2.hdf5", p2s_map=phonon_p2s_map)
        _check_fc2_shape(ph3py, fc2)
        ph3py.fc2 = fc2
        read_fc2 = True
        if log_level:
            print('fc2 was read from "fc2.hdf5".')
    elif os.path.isfile("FORCES_FC2"):
        disp_filename = None
        if os.path.isfile("disp_fc2.yaml"):
            if ph3py_yaml is None:
                disp_filename = "disp_fc2.yaml"
            elif ph3py_yaml.phonon_dataset is None:
                disp_filename = "disp_fc2.yaml"
        _set_dataset_for_fc2(
            ph3py,
            ph3py_yaml,
            "FORCES_FC2",
            disp_filename,
            "phonon_fc2",
            log_level,
        )
    elif (
        ph3py_yaml is not None
        and ph3py_yaml.phonon_dataset is not None
        and forces_in_dataset(ph3py_yaml.phonon_dataset)
    ):
        _set_dataset_for_fc2(
            ph3py,
            ph3py_yaml,
            None,
            None,
            "phonon_fc2",
            log_level,
        )
    elif (
        ph3py_yaml is not None
        and ph3py_yaml.dataset is not None
        and forces_in_dataset(ph3py_yaml.dataset)
    ):
        _set_dataset_for_fc2(
            ph3py,
            ph3py_yaml,
            None,
            None,
            "fc2",
            log_level,
        )
    elif os.path.isfile("FORCES_FC3"):
        # suppose fc3.hdf5 is read but fc2.hdf5 doesn't exist.
        disp_filename = None
        if os.path.isfile("disp_fc3.yaml"):
            if ph3py_yaml is None:
                disp_filename = "disp_fc3.yaml"
            elif ph3py_yaml.dataset is None:
                disp_filename = "disp_fc3.yaml"
        _set_dataset_for_fc2(
            ph3py,
            ph3py_yaml,
            "FORCES_FC3",
            disp_filename,
            "fc2",
            log_level,
        )
    return read_fc2


def _set_dataset_for_fc3(
    ph3py: Phono3py,
    ph3py_yaml: Optional[Phono3pyYaml],
    force_filename,
    disp_filename,
    phono3py_yaml_filename,
    cutoff_pair_distance,
    log_level,
):
    ph3py.dataset = parse_forces(
        ph3py,
        ph3py_yaml=ph3py_yaml,
        cutoff_pair_distance=cutoff_pair_distance,
        force_filename=force_filename,
        disp_filename=disp_filename,
        phono3py_yaml_filename=phono3py_yaml_filename,
        fc_type="fc3",
        log_level=log_level,
    )


def _set_dataset_for_fc2(
    ph3py: Phono3py,
    ph3py_yaml: Optional[Phono3pyYaml],
    force_filename,
    disp_filename,
    fc_type,
    log_level,
):
    dataset = parse_forces(
        ph3py,
        ph3py_yaml=ph3py_yaml,
        force_filename=force_filename,
        disp_filename=disp_filename,
        fc_type=fc_type,
        log_level=log_level,
    )
    if fc_type == "phonon_fc2":
        ph3py.phonon_dataset = dataset
    else:
        ph3py.dataset = dataset


def _check_fc2_shape(ph3py: Phono3py, fc2, filename="fc2.hdf5"):
    if ph3py.phonon_supercell_matrix is None:
        smat = ph3py.supercell_matrix
    else:
        smat = ph3py.phonon_supercell_matrix
    _check_fc_shape(ph3py, fc2, smat, filename)


def _check_fc3_shape(ph3py: Phono3py, fc3, filename="fc3.hdf5"):
    smat = ph3py.supercell_matrix
    _check_fc_shape(ph3py, fc3, smat, filename)


def _check_fc_shape(ph3py: Phono3py, fc, smat, filename):
    if len(ph3py.unitcell) * determinant(smat) != fc.shape[1]:
        msg = (
            f'Supercell size mismatch between "{filename}" and supercell matrix {smat}.'
        )
        raise RuntimeError(msg)



"""phono3py_yaml reader and writer."""
# Copyright (C) 2016 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Optional

import numpy as np
from phonopy.interface.phonopy_yaml import (
    PhonopyYaml,
    PhonopyYamlDumper,
    PhonopyYamlLoader,
    load_yaml,
    phonopy_yaml_property_factory,
)

if TYPE_CHECKING:
    from phono3py import Phono3py

from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive, Supercell
from phonopy.structure.symmetry import Symmetry


@dataclasses.dataclass
class Phono3pyYamlData:
    """PhonopyYaml data structure."""

    configuration: dict
    calculator: str
    physical_units: dict
    unitcell: Optional[PhonopyAtoms] = None
    primitive: Optional[Primitive] = None
    supercell: Optional[Supercell] = None
    dataset: Optional[dict] = None
    supercell_matrix: Optional[np.ndarray] = None
    primitive_matrix: Optional[np.ndarray] = None
    nac_params: Optional[dict] = None
    force_constants: Optional[np.ndarray] = None
    symmetry: Optional[Symmetry] = None  # symmetry of supercell
    frequency_unit_conversion_factor: Optional[float] = None
    version: Optional[str] = None
    command_name: str = "phono3py"

    phonon_supercell_matrix: Optional[np.ndarray] = None
    phonon_dataset: Optional[dict] = None
    phonon_supercell: Optional[Supercell] = None
    phonon_primitive: Optional[Primitive] = None


class Phono3pyYamlLoader(PhonopyYamlLoader):
    """Phono3pyYaml loader."""

    def __init__(
        self, yaml_data, configuration=None, calculator=None, physical_units=None
    ):
        """Init method.

        Parameters
        ----------
        yaml_data : dict

        """
        self._yaml = yaml_data
        self._data = Phono3pyYamlData(
            configuration=configuration,
            calculator=calculator,
            physical_units=physical_units,
        )

    def parse(self):
        """Yaml dict is parsed. See docstring of this class."""
        super().parse()
        self._parse_fc3_dataset()
        return self

    def _parse_all_cells(self):
        """Parse all cells.

        This method override PhonopyYaml._parse_all_cells.

        """
        super()._parse_all_cells()
        if "phonon_primitive_cell" in self._yaml:
            self._data.phonon_primitive = self._parse_cell(
                self._yaml["phonon_primitive_cell"]
            )
        if "phonon_supercell" in self._yaml:
            self._data.phonon_supercell = self._parse_cell(
                self._yaml["phonon_supercell"]
            )
        if "phonon_supercell_matrix" in self._yaml:
            self._data.phonon_supercell_matrix = np.array(
                self._yaml["phonon_supercell_matrix"], dtype="intc", order="C"
            )

    def _parse_dataset(self):
        """Parse phonon_dataset.

        This method override PhonopyYaml._parse_dataset.

        key="phonon_displacements" at v2.2 or later.
        key="displacements" at older version than v2.2.

        """
        self._data.phonon_dataset = self._get_dataset(
            self._data.phonon_supercell, key="phonon_displacements"
        )
        if self._data.phonon_dataset is None:  # key="displacements"
            self._data.phonon_dataset = self._get_dataset(self._data.phonon_supercell)

    def _parse_fc3_dataset(self):
        """Parse force dataset for fc3.

        'duplicates' can be either dict (<v1.21) or list in phono3py.yaml.
        From v1.21, it was changed to list of list because
        dict with a key of int type is not allowed in JSON.

        "displacements" gives type-II fc3 dataset at v2.2 or later, although
        it gave phonon-dataset at versions older than v2.2.

        """
        dataset = None
        if "displacement_pairs" in self._yaml:
            disp = self._yaml["displacement_pairs"][0]
            if type(disp) is dict:  # type1
                dataset = self._parse_fc3_dataset_type1(len(self._data.supercell))
            elif type(disp) is list:  # type2
                if "displacement" in disp[0]:
                    dataset = self._parse_force_sets_type2()
        if "displacement_pair_info" in self._yaml:
            info_yaml = self._yaml["displacement_pair_info"]
            if "cutoff_pair_distance" in info_yaml:
                dataset["cutoff_distance"] = info_yaml["cutoff_pair_distance"]
            if "duplicated_supercell_ids" in info_yaml:
                dataset["duplicates"] = info_yaml["duplicated_supercell_ids"]
        self._data.dataset = dataset

        # This case should work only for v2.2 or later.
        if self._data.dataset is None and "displacements" in self._yaml:
            self._data.dataset = self._get_dataset(self._data.supercell)

    def _parse_fc3_dataset_type1(self, natom):
        """Parse fc3 type1-dataset."""
        dataset = {"natom": natom, "first_atoms": []}
        disp2_id = len(self._yaml["displacement_pairs"])
        for disp1_id, d1 in enumerate(self._yaml["displacement_pairs"]):
            data1 = {
                "number": d1["atom"] - 1,
                "displacement": np.array(d1["displacement"], dtype="double"),
                "second_atoms": [],
            }
            if "displacement_id" in d1:
                d1_id = d1["displacement_id"]
                if disp1_id + 1 != d1_id:
                    msg = f"{d1_id} != {disp1_id + 1}. Dataset may be broken."
                    raise RuntimeError(msg)
                data1["id"] = d1_id
            if "forces" in d1:
                data1["forces"] = np.array(d1["forces"], dtype="double", order="C")
            d2_list = d1.get("paired_with")
            if d2_list is None:  # backward compatibility
                d2_list = d1.get("second_atoms")
            for d2 in d2_list:
                if "forces" in d2:
                    disp2_id = self._parse_fc3_dataset_type1_with_forces(
                        data1, d2, disp2_id
                    )
                else:
                    disp2_id = self._parse_fc3_dataset_type1_without_forces(
                        data1, d2, disp2_id
                    )
            dataset["first_atoms"].append(data1)
        return dataset

    def _parse_fc3_dataset_type1_with_forces(self, data1, d2, disp2_id):
        """Parse fc3 type1-dataset that has forces in it.

        One displacement couples with one force-set in yaml.

        """
        second_atom_dict = {
            "number": d2["atom"] - 1,
            "displacement": np.array(d2["displacement"], dtype="double"),
            "forces": np.array(d2["forces"], dtype="double", order="C"),
        }
        if "displacement_id" in d2:
            d2_id = d2["displacement_id"]
            if disp2_id + 1 != d2_id:
                msg = f"{d2_id} != {disp2_id + 1}. Dataset may be broken."
                raise RuntimeError(msg)
            second_atom_dict["id"] = d2_id
        if "pair_distance" in d2:
            second_atom_dict["pair_distance"] = d2["pair_distance"]
        disp2_id += 1
        data1["second_atoms"].append(second_atom_dict)
        return disp2_id

    def _parse_fc3_dataset_type1_without_forces(self, data1, d2, disp2_id):
        """Parse fc3 type1-dataset that doesn't have forces in it.

        Displacements are stored in `displacements` as a list.

        """
        disps = [
            {
                "number": d2["atom"] - 1,
                "displacement": np.array(disp, dtype="double"),
            }
            for disp in d2["displacements"]
        ]
        if "pair_distance" in d2:
            for d2_dict in disps:
                d2_dict["pair_distance"] = d2["pair_distance"]
        if "included" in d2:
            for d2_dict in disps:
                d2_dict["included"] = d2["included"]
        if "displacement_ids" in d2:
            for d2_id, d2_dict in zip(d2["displacement_ids"], disps):
                if disp2_id + 1 != d2_id:
                    msg = f"{d2_id} != {disp2_id + 1}. Dataset may be broken."
                    raise RuntimeError(msg)
                d2_dict["id"] = d2_id
                disp2_id += 1
        else:
            disp2_id += len(disps)
        data1["second_atoms"] += disps
        return disp2_id


class Phono3pyYamlDumper(PhonopyYamlDumper):
    """Phono3pyYaml dumper."""

    _default_dumper_settings = {
        "force_sets": False,
        "displacements": True,
        "force_constants": False,
        "born_effective_charge": True,
        "dielectric_constant": True,
    }

    def __init__(self, data: Phono3pyYamlData, dumper_settings=None):
        """Init method."""
        self._data = data
        self._init_dumper_settings(dumper_settings)

    def _cell_info_yaml_lines(self):
        """Get YAML lines for information of cells.

        This method override PhonopyYaml._cell_info_yaml_lines.

        """
        lines = super()._cell_info_yaml_lines()
        lines += self._supercell_matrix_yaml_lines(
            self._data.phonon_supercell_matrix, "phonon_supercell_matrix"
        )
        lines += self._primitive_yaml_lines(
            self._data.phonon_primitive, "phonon_primitive_cell"
        )
        lines += self._phonon_supercell_yaml_lines()
        return lines

    def _phonon_supercell_matrix_yaml_lines(self):
        lines = []
        if self._data.phonon_supercell_matrix is not None:
            lines.append("phonon_supercell_matrix:")
            for v in self._data.supercell_matrix:
                lines.append("- [ %3d, %3d, %3d ]" % tuple(v))
            lines.append("")
        return lines

    def _phonon_supercell_yaml_lines(self):
        lines = []
        if self._data.phonon_supercell is not None:
            s2p_map = getattr(self._data.phonon_primitive, "s2p_map", None)
            lines += self._cell_yaml_lines(
                self._data.phonon_supercell, "phonon_supercell", s2p_map
            )
            lines.append("")
        return lines

    def _nac_yaml_lines(self):
        """Get YAML lines for parameters of non-analytical term correction.

        This method override PhonopyYaml._nac_yaml_lines.

        """
        if self._data.phonon_primitive is not None:
            return self._nac_yaml_lines_given_symbols(
                self._data.phonon_primitive.symbols
            )
        else:
            return self._nac_yaml_lines_given_symbols(self._data.primitive.symbols)

    def _displacements_yaml_lines(self, with_forces=False, key="displacements"):
        """Get YAML lines for phonon_dataset and dataset.

        This method override PhonopyYaml._displacements_yaml_lines.
        PhonopyYaml._displacements_yaml_lines_2types is written
        to be also used by Phono3pyYaml.

        """
        lines = []
        if self._data.phonon_supercell_matrix is not None:
            lines += self._displacements_yaml_lines_2types(
                self._data.phonon_dataset, with_forces=with_forces, key=f"phonon_{key}"
            )
        lines += self._displacements_yaml_lines_2types(
            self._data.dataset, with_forces=with_forces, key=key
        )
        return lines

    def _displacements_yaml_lines_type1(
        self, dataset, with_forces=False, key="displacements"
    ):
        """Get YAML lines for type1 phonon_dataset and dataset.

        This method override PhonopyYaml._displacements_yaml_lines_type1.
        PhonopyYaml._displacements_yaml_lines_2types calls
        Phono3pyYaml._displacements_yaml_lines_type1.

        """
        return displacements_yaml_lines_type1(dataset, with_forces=with_forces, key=key)


class Phono3pyYaml(PhonopyYaml):
    """phono3py.yaml reader and writer.

    Details are found in the docstring of PhonopyYaml.
    The common usages are as follows:

    1. Set phono3py instance.
        p3yml = Phono3pyYaml()
        p3yml.set_phonon_info(phono3py_instance)
    2. Read phono3py.yaml file.
        p3yml = Phono3pyYaml()
        p3yml.read(filename)
    3. Parse yaml dict of phono3py.yaml.
        with open("phono3py.yaml", 'r') as f:
            p3yml.yaml_data = yaml.load(f, Loader=yaml.CLoader)
            p3yml.parse()
    4. Save stored data in Phono3pyYaml instance into a text file in yaml.
        with open(filename, 'w') as w:
            w.write(str(ph3py_yaml))

    """

    default_filenames = ("phono3py_disp.yaml", "phono3py.yaml")
    command_name = "phono3py"

    configuration = phonopy_yaml_property_factory("configuration")
    calculator = phonopy_yaml_property_factory("calculator")
    physical_units = phonopy_yaml_property_factory("physical_units")
    unitcell = phonopy_yaml_property_factory("unitcell")
    primitive = phonopy_yaml_property_factory("primitive")
    supercell = phonopy_yaml_property_factory("supercell")
    dataset = phonopy_yaml_property_factory("dataset")
    supercell_matrix = phonopy_yaml_property_factory("supercell_matrix")
    primitive_matrix = phonopy_yaml_property_factory("primitive_matrix")
    nac_params = phonopy_yaml_property_factory("nac_params")
    force_constants = phonopy_yaml_property_factory("force_constants")
    symmetry = phonopy_yaml_property_factory("symmetry")
    frequency_unit_conversion_factor = phonopy_yaml_property_factory(
        "frequency_unit_conversion_factor"
    )
    version = phonopy_yaml_property_factory("version")

    phonon_supercell_matrix = phonopy_yaml_property_factory("phonon_supercell_matrix")
    phonon_dataset = phonopy_yaml_property_factory("phonon_dataset")
    phonon_supercell = phonopy_yaml_property_factory("phonon_supercell")
    phonon_primitive = phonopy_yaml_property_factory("phonon_primitive")

    def __init__(
        self, configuration=None, calculator=None, physical_units=None, settings=None
    ):
        """Init method."""
        self._data = Phono3pyYamlData(
            configuration=configuration,
            calculator=calculator,
            physical_units=physical_units,
        )
        self._dumper_settings = settings

    def __str__(self):
        """Return string text of yaml output."""
        ph3yml_dumper = Phono3pyYamlDumper(
            self._data, dumper_settings=self._dumper_settings
        )
        return "\n".join(ph3yml_dumper.get_yaml_lines())

    def read(self, filename):
        """Read Phono3pyYaml file."""
        self._data = read_phono3py_yaml(
            filename,
            configuration=self._data.configuration,
            calculator=self._data.calculator,
            physical_units=self._data.physical_units,
        )
        return self

    def set_phonon_info(self, phono3py: "Phono3py"):
        """Store data in Phono3py instance in this instance."""
        super().set_phonon_info(phono3py)
        self._data.phonon_supercell_matrix = phono3py.phonon_supercell_matrix
        self._data.phonon_dataset = phono3py.phonon_dataset
        self._data.phonon_primitive = phono3py.phonon_primitive
        self._data.phonon_supercell = phono3py.phonon_supercell


def displacements_yaml_lines_type1(dataset, with_forces=False, key="displacements"):
    """Get YAML lines for type1 phonon_dataset and dataset.

    This is a function but not class method because used by other function.

    """
    id_offset = len(dataset["first_atoms"])

    if "second_atoms" in dataset["first_atoms"][0]:
        lines = ["displacement_pairs:"]
    else:
        lines = [f"{key}:"]
    for disp1_id, d1 in enumerate(dataset["first_atoms"]):
        lines.append("- atom: %4d" % (d1["number"] + 1))
        lines.append("  displacement:")
        lines.append("    [ %19.16f, %19.16f, %19.16f ]" % tuple(d1["displacement"]))
        if "id" in d1:
            assert disp1_id + 1 == d1["id"]
        lines.append(f"  displacement_id: {disp1_id + 1}")
        if with_forces and "forces" in d1:
            lines.append("  forces:")
            for v in d1["forces"]:
                lines.append("  - [ %19.16f, %19.16f, %19.16f ]" % tuple(v))
        if "second_atoms" in d1:
            ret_lines, id_offset = _second_displacements_yaml_lines(
                d1["second_atoms"], id_offset, with_forces=with_forces
            )
            lines += ret_lines
    lines.append("")

    if "second_atoms" in dataset["first_atoms"][0]:
        lines += _displacements_yaml_lines_type1_info(dataset)

    return lines


def _displacements_yaml_lines_type1_info(dataset):
    """Return lines of displacement-pair summary."""
    n_single = len(dataset["first_atoms"])
    n_pair = 0
    n_included = 0
    for d1 in dataset["first_atoms"]:
        n_pair += len(d1["second_atoms"])
        for d2 in d1["second_atoms"]:
            if "included" not in d2:
                n_included += 1
            elif d2["included"]:
                n_included += 1

    lines = []
    lines.append("displacement_pair_info:")
    if "cutoff_distance" in dataset:
        lines.append("  cutoff_pair_distance: %11.8f" % dataset["cutoff_distance"])
    lines.append("  number_of_singles: %d" % n_single)
    lines.append("  number_of_pairs: %d" % n_pair)
    if "cutoff_distance" in dataset:
        lines.append("  number_of_pairs_in_cutoff: %d" % n_included)

    # 'duplicates' is dict, but written as a list of list in yaml.
    # See the docstring of _parse_fc3_dataset for the reason.
    if "duplicates" in dataset and dataset["duplicates"]:
        lines.append("  duplicated_supercell_ids: " "# 0 means perfect supercell")
        # Backward compatibility for dict type
        if type(dataset["duplicates"]) is dict:
            for disp1_id, j in dataset["duplicates"].items():
                lines.append("  - [ %d, %d ]" % (int(disp1_id), j))
        else:
            for disp1_id, j in dataset["duplicates"]:
                lines.append("  - [ %d, %d ]" % (disp1_id, j))
        lines.append("")

    return lines


def _second_displacements_yaml_lines(dataset2, id_offset, with_forces=False):
    lines = []
    disp2_id = id_offset
    # lines.append("  second_atoms:")
    lines.append("  paired_with:")
    numbers = np.array([d["number"] for d in dataset2])
    unique_numbers = np.unique(numbers)
    for i in unique_numbers:
        indices_eq_i = np.sort(np.where(numbers == i)[0])
        if with_forces and "forces" in dataset2[indices_eq_i[0]]:
            for j in indices_eq_i:
                disp2_id += 1
                lines.append("  - atom: %4d" % (i + 1))
                lines.append("    pair_distance: %.8f" % dataset2[j]["pair_distance"])
                lines.append("    displacement:")
                lines.append(
                    "      [ %19.16f, %19.16f, %19.16f ]"
                    % tuple(dataset2[j]["displacement"])
                )

                if "id" in dataset2[j]:
                    assert dataset2[j]["id"] == disp2_id
                lines.append("    displacement_id: %d" % disp2_id)

                lines.append("    forces:")
                for v in dataset2[j]["forces"]:
                    lines.append("    - [ %19.16f, %19.16f, %19.16f ]" % tuple(v))
        else:
            lines.append("  - atom: %4d" % (i + 1))
            lines.append(
                "    pair_distance: %.8f" % dataset2[indices_eq_i[0]]["pair_distance"]
            )
            if "included" in dataset2[indices_eq_i[0]]:
                included = dataset2[indices_eq_i[0]]["included"]
                lines.append("    included: %s" % ("true" if included else "false"))
            disp_ids = []
            lines.append("    displacements:")
            for j in indices_eq_i:
                disp2_id += 1
                d = tuple(dataset2[j]["displacement"])
                lines.append("    - [ %19.16f, %19.16f, %19.16f ]" % d)
                if "id" in dataset2[j]:
                    assert dataset2[j]["id"] == disp2_id
                disp_ids.append(disp2_id)
            lines.append(
                "    displacement_ids: [ %s ]" % ", ".join(["%d" % j for j in disp_ids])
            )

    return lines, disp2_id


def read_phono3py_yaml(
    filename, configuration=None, calculator=None, physical_units=None
) -> Phono3pyYamlData:
    """Read phono3py.yaml like file."""
    yaml_data = load_yaml(filename)
    if type(yaml_data) is str:
        msg = f'Could not load "{filename}" properly.'
        raise TypeError(msg)
    return load_phono3py_yaml(
        yaml_data,
        configuration=configuration,
        calculator=calculator,
        physical_units=physical_units,
    )


def load_phono3py_yaml(
    yaml_data, configuration=None, calculator=None, physical_units=None
) -> Phono3pyYamlData:
    """Return Phono3pyYamlData instance loading yaml data.

    Parameters
    -----------
    yaml_data : dict

    """
    ph3yml_loader = Phono3pyYamlLoader(
        yaml_data,
        configuration=configuration,
        calculator=calculator,
        physical_units=physical_units,
    )
    ph3yml_loader.parse()
    return ph3yml_loader.data



"""Tests of PhonopyYaml."""
from io import StringIO
from pathlib import Path

import numpy as np
import yaml
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import isclose
from phonopy.structure.dataset import get_displacements_and_forces

from phono3py import Phono3py
from phono3py.interface.phono3py_yaml import Phono3pyYaml, load_phono3py_yaml

cwd = Path(__file__).parent


def test_read_phono3py_yaml():
    """Test to parse phono3py.yaml like file."""
    filename = cwd / ".." / "phono3py_params_NaCl111.yaml"
    cell = _get_unitcell(filename)
    cell_ref = PhonopyAtoms(
        cell=[[5.64056, 0.0, 0.0], [0.0, 5.64056, 0.0], [0.0, 0.0, 5.64056]],
        scaled_positions=[
            [0.0, 0.0, 0.0],
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 0.5, 0.0],
            [0.5, 0.5, 0.5],
            [0.5, 0.0, 0.0],
            [0.0, 0.5, 0.0],
            [0.0, 0.0, 0.5],
        ],
        numbers=[11, 11, 11, 11, 17, 17, 17, 17],
    )
    assert isclose(cell, cell_ref)


def test_write_phono3py_yaml(nacl_pbe: Phono3py):
    """Test Phono3pyYaml.set_phonon_info, __str__, yaml_data, parse."""
    phonon3 = nacl_pbe
    ph3py_yaml = Phono3pyYaml(calculator="vasp")
    ph3py_yaml.set_phonon_info(phonon3)
    ph3py_yaml_test = Phono3pyYaml()
    ph3py_yaml_test._data = load_phono3py_yaml(
        yaml.safe_load(StringIO(str(ph3py_yaml))), calculator=ph3py_yaml.calculator
    )
    assert isclose(ph3py_yaml.primitive, ph3py_yaml_test.primitive)
    assert isclose(ph3py_yaml.unitcell, ph3py_yaml_test.unitcell)
    assert isclose(ph3py_yaml.supercell, ph3py_yaml_test.supercell)
    assert ph3py_yaml.version == ph3py_yaml_test.version
    np.testing.assert_allclose(
        ph3py_yaml.supercell_matrix, ph3py_yaml_test.supercell_matrix, atol=1e-8
    )
    np.testing.assert_allclose(
        ph3py_yaml.primitive_matrix, ph3py_yaml_test.primitive_matrix, atol=1e-8
    )


def test_write_phono3py_yaml_extra_NaCl222(nacl_pbe: Phono3py):
    """Test Phono3pyYaml.set_phonon_info, __str__, yaml_data, parse.

    settings parameter controls amount of yaml output. In this test,
    more data than the default are dumped and those are tested.

    """
    _test_write_phono3py_yaml_extra(nacl_pbe)


def test_write_phono3py_yaml_extra_Si111_222(si_pbesol_111_222_fd: Phono3py):
    """Test Phono3pyYaml.set_phonon_info, __str__, yaml_data, parse.

    settings parameter controls amount of yaml output. In this test,
    more data than the default are dumped and those are tested.

    """
    _test_write_phono3py_yaml_extra(si_pbesol_111_222_fd)


def _test_write_phono3py_yaml_extra(phonon3: Phono3py):
    settings = {
        "force_sets": True,
        "displacements": True,
        "force_constants": True,
        "born_effective_charge": True,
        "dielectric_constant": True,
    }
    ph3py_yaml = Phono3pyYaml(calculator="vasp", settings=settings)
    ph3py_yaml.set_phonon_info(phonon3)
    ph3py_yaml_test = Phono3pyYaml()
    ph3py_yaml_test._data = load_phono3py_yaml(
        yaml.safe_load(StringIO(str(ph3py_yaml))), calculator=ph3py_yaml.calculator
    )
    np.testing.assert_allclose(
        ph3py_yaml.force_constants, ph3py_yaml_test.force_constants, atol=1e-8
    )

    if ph3py_yaml.nac_params is not None:
        np.testing.assert_allclose(
            ph3py_yaml.nac_params["born"], ph3py_yaml_test.nac_params["born"], atol=1e-8
        )
        np.testing.assert_allclose(
            ph3py_yaml.nac_params["dielectric"],
            ph3py_yaml_test.nac_params["dielectric"],
            atol=1e-8,
        )
        np.testing.assert_allclose(
            ph3py_yaml.nac_params["factor"],
            ph3py_yaml_test.nac_params["factor"],
            atol=1e-8,
        )

    disps, forces = get_displacements_and_forces(ph3py_yaml.dataset)
    disps_test, forces_test = get_displacements_and_forces(ph3py_yaml_test.dataset)
    np.testing.assert_allclose(forces, forces_test, atol=1e-8)
    np.testing.assert_allclose(disps, disps_test, atol=1e-8)

    if ph3py_yaml.phonon_dataset is not None:
        disps, forces = get_displacements_and_forces(ph3py_yaml.phonon_dataset)
        disps_test, forces_test = get_displacements_and_forces(
            ph3py_yaml_test.phonon_dataset
        )
        np.testing.assert_allclose(forces, forces_test, atol=1e-8)
        np.testing.assert_allclose(disps, disps_test, atol=1e-8)


def _get_unitcell(filename):
    phpy_yaml = Phono3pyYaml().read(filename)
    return phpy_yaml.unitcell



"""Test for Conductivity_RTA.py."""
import numpy as np

# first list is k_P, second list is k_C
si_pbesol_kappa_P_RTA = [108.723, 108.723, 108.723, 0.000, 0.000, 0.000]
si_pbesol_kappa_C = [0.167, 0.167, 0.167, 0.000, 0.000, 0.000]

si_pbesol_kappa_P_RTA_iso = [98.008, 98.008, 98.008, 0.000, 0.000, 0.000]
si_pbesol_kappa_C_iso = [0.177, 0.177, 0.177, 0.000, 0.000, 0.000]

si_pbesol_kappa_P_RTA_with_sigmas = [110.534, 110.534, 110.534, 0, 0, 0]
si_pbesol_kappa_C_with_sigmas = [0.163, 0.163, 0.163, 0.000, 0.000, 0.000]

si_pbesol_kappa_P_RTA_with_sigmas_iso = [97.268, 97.268, 97.268, 0, 0, 0]
si_pbesol_kappa_C_with_sigmas_iso = [0.179, 0.179, 0.179, 0.000, 0.000, 0.000]

si_pbesol_kappa_P_RTA_si_nosym = [39.325, 39.323, 39.496, -0.004, 0.020, 0.018]

si_pbesol_kappa_C_si_nosym = [0.009, 0.009, 0.009, 0.000, 0.000, 0.000]

si_pbesol_kappa_P_RTA_si_nomeshsym = [39.411, 39.411, 39.411, 0, 0, 0]
si_pbesol_kappa_C_si_nomeshsym = [0.009, 0.009, 0.009, 0.000, 0.000, 0.000]

nacl_pbe_kappa_P_RTA = [7.753, 7.753, 7.753, 0.000, 0.000, 0.000]
nacl_pbe_kappa_C = [0.081, 0.081, 0.081, 0.000, 0.000, 0.000]

nacl_pbe_kappa_RTA_with_sigma = [7.742, 7.742, 7.742, 0, 0, 0]  # old value 7.71913708
nacl_pbe_kappa_C_with_sigma = [0.081, 0.081, 0.081, 0.000, 0.000, 0.000]

aln_lda_kappa_P_RTA = [203.304, 203.304, 213.003, 0, 0, 0]
aln_lda_kappa_C = [0.084, 0.084, 0.037, 0, 0, 0]

aln_lda_kappa_P_RTA_with_sigmas = [213.820000, 213.820000, 224.800000, 0, 0, 0]
aln_lda_kappa_C_with_sigmas = [0.084, 0.084, 0.036, 0, 0, 0]


def test_kappa_RTA_si(si_pbesol):
    """Test RTA by Si."""
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol, [9, 9, 9])
    np.testing.assert_allclose(si_pbesol_kappa_P_RTA, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C, kappa_C, atol=0.02)


def test_kappa_RTA_si_full_pp(si_pbesol):
    """Test RTA with full-pp by Si."""
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol, [9, 9, 9], is_full_pp=True)
    np.testing.assert_allclose(si_pbesol_kappa_P_RTA, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C, kappa_C, atol=0.02)


def test_kappa_RTA_si_iso(si_pbesol):
    """Test RTA with isotope scattering by Si."""
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol, [9, 9, 9], is_isotope=True)
    np.testing.assert_allclose(si_pbesol_kappa_P_RTA_iso, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C_iso, kappa_C, atol=0.02)


def test_kappa_RTA_si_with_sigma(si_pbesol):
    """Test RTA with smearing method by Si."""
    si_pbesol.sigmas = [
        0.1,
    ]
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol, [9, 9, 9])
    np.testing.assert_allclose(si_pbesol_kappa_P_RTA_with_sigmas, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C_with_sigmas, kappa_C, atol=0.02)
    si_pbesol.sigmas = None


def test_kappa_RTA_si_with_sigma_full_pp(si_pbesol):
    """Test RTA with smearing method and full-pp by Si."""
    si_pbesol.sigmas = [
        0.1,
    ]
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol, [9, 9, 9], is_full_pp=True)
    np.testing.assert_allclose(si_pbesol_kappa_P_RTA_with_sigmas, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C_with_sigmas, kappa_C, atol=0.02)
    si_pbesol.sigmas = None


def test_kappa_RTA_si_with_sigma_iso(si_pbesol):
    """Test RTA with smearing method and isotope scattering by Si."""
    si_pbesol.sigmas = [
        0.1,
    ]
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol, [9, 9, 9], is_isotope=True)
    np.testing.assert_allclose(
        si_pbesol_kappa_P_RTA_with_sigmas_iso, kappa_P_RTA, atol=0.5
    )
    np.testing.assert_allclose(si_pbesol_kappa_C_with_sigmas_iso, kappa_C, atol=0.02)
    si_pbesol.sigmas = None


def test_kappa_RTA_si_compact_fc(si_pbesol_compact_fc):
    """Test RTA with compact-fc by Si."""
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol_compact_fc, [9, 9, 9])
    np.testing.assert_allclose(si_pbesol_kappa_P_RTA, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C, kappa_C, atol=0.02)


def test_kappa_RTA_si_nosym(si_pbesol, si_pbesol_nosym):
    """Test RTA without considering symmetry by Si."""
    si_pbesol_nosym.fc2 = si_pbesol.fc2
    si_pbesol_nosym.fc3 = si_pbesol.fc3
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol_nosym, [4, 4, 4])
    kappa_P_RTA_r = kappa_P_RTA.reshape(-1, 3).sum(axis=1)
    kappa_C_r = kappa_C.reshape(-1, 3).sum(axis=1)
    kappa_P_ref = np.reshape(si_pbesol_kappa_P_RTA_si_nosym, (-1, 3)).sum(axis=1)
    kappa_C_ref = np.reshape(si_pbesol_kappa_C_si_nosym, (-1, 3)).sum(axis=1)
    np.testing.assert_allclose(kappa_P_ref / 3, kappa_P_RTA_r / 3, atol=0.8)
    np.testing.assert_allclose(kappa_C_ref / 3, kappa_C_r / 3, atol=0.02)


def test_kappa_RTA_si_nomeshsym(si_pbesol, si_pbesol_nomeshsym):
    """Test RTA without considering mesh symmetry by Si."""
    si_pbesol_nomeshsym.fc2 = si_pbesol.fc2
    si_pbesol_nomeshsym.fc3 = si_pbesol.fc3
    kappa_P_RTA, kappa_C = _get_kappa_RTA(si_pbesol_nomeshsym, [4, 4, 4])
    np.testing.assert_allclose(
        si_pbesol_kappa_P_RTA_si_nomeshsym, kappa_P_RTA, atol=1.0
    )
    np.testing.assert_allclose(si_pbesol_kappa_C_si_nomeshsym, kappa_C, atol=0.02)


def test_kappa_RTA_si_N_U(si_pbesol):
    """Test RTA with N and U scatterings by Si."""
    ph3 = si_pbesol
    mesh = [4, 4, 4]
    is_N_U = True
    ph3.mesh_numbers = mesh
    ph3.init_phph_interaction()
    ph3.run_thermal_conductivity(
        temperatures=[
            300,
        ],
        is_N_U=is_N_U,
        conductivity_type="wigner",
    )
    gN, gU = ph3.thermal_conductivity.get_gamma_N_U()

    gN_ref = [
        0.00000000,
        0.00000000,
        0.00000000,
        0.07402084,
        0.07402084,
        0.07402084,
        0.00078535,
        0.00078535,
        0.00917995,
        0.02178049,
        0.04470075,
        0.04470075,
        0.00173337,
        0.00173337,
        0.01240191,
        0.00198981,
        0.03165195,
        0.03165195,
        0.00224713,
        0.00224713,
        0.00860026,
        0.03083611,
        0.03083611,
        0.02142118,
        0.00277534,
        0.00330170,
        0.02727451,
        0.00356415,
        0.01847744,
        0.01320643,
        0.00155072,
        0.00365611,
        0.01641919,
        0.00650083,
        0.02576069,
        0.01161589,
        0.00411969,
        0.00411969,
        0.00168211,
        0.00168211,
        0.01560092,
        0.01560092,
        0.00620091,
        0.00620091,
        0.03764912,
        0.03764912,
        0.02668523,
        0.02668523,
    ]
    gU_ref = [
        0.00000000,
        0.00000000,
        0.00000000,
        0.00000000,
        0.00000000,
        0.00000000,
        0.00015178,
        0.00015178,
        0.00076936,
        0.00727539,
        0.00113112,
        0.00113112,
        0.00022696,
        0.00022696,
        0.00072558,
        0.00000108,
        0.00021968,
        0.00021968,
        0.00079397,
        0.00079397,
        0.00111068,
        0.00424761,
        0.00424761,
        0.00697760,
        0.00221593,
        0.00259510,
        0.01996296,
        0.00498962,
        0.01258375,
        0.00513825,
        0.00148802,
        0.00161955,
        0.01589219,
        0.00646134,
        0.00577275,
        0.00849711,
        0.00313208,
        0.00313208,
        0.00036610,
        0.00036610,
        0.01135335,
        0.01135335,
        0.00000000,
        0.00000000,
        0.00000000,
        0.00000000,
        0.00000000,
        0.00000000,
    ]

    # print(np.sum(gN), np.sum(gU))
    np.testing.assert_allclose(
        np.sum([gN_ref, gU_ref], axis=0), gN.ravel() + gU.ravel(), atol=1e-2
    )
    np.testing.assert_allclose(gN_ref, gN.ravel(), atol=1e-2)
    np.testing.assert_allclose(gU_ref, gU.ravel(), atol=1e-2)


def test_kappa_RTA_nacl(nacl_pbe):
    """Test RTA by NaCl."""
    kappa_P_RTA, kappa_C = _get_kappa_RTA(nacl_pbe, [9, 9, 9])
    np.testing.assert_allclose(nacl_pbe_kappa_P_RTA, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(nacl_pbe_kappa_C, kappa_C, atol=0.02)


def test_kappa_RTA_nacl_with_sigma(nacl_pbe):
    """Test RTA with smearing method by NaCl."""
    nacl_pbe.sigmas = [
        0.1,
    ]
    nacl_pbe.sigma_cutoff = 3
    kappa_P_RTA, kappa_C = _get_kappa_RTA(nacl_pbe, [9, 9, 9])
    np.testing.assert_allclose(nacl_pbe_kappa_RTA_with_sigma, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(nacl_pbe_kappa_C_with_sigma, kappa_C, atol=0.02)
    nacl_pbe.sigmas = None
    nacl_pbe.sigma_cutoff = None


def test_kappa_RTA_aln(aln_lda):
    """Test RTA by AlN."""
    kappa_P_RTA, kappa_C = _get_kappa_RTA(aln_lda, [7, 7, 5])
    np.testing.assert_allclose(aln_lda_kappa_P_RTA, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(aln_lda_kappa_C, kappa_C, atol=0.02)


def test_kappa_RTA_aln_with_sigma(aln_lda):
    """Test RTA with smearing method by AlN."""
    aln_lda.sigmas = [
        0.1,
    ]
    aln_lda.sigma_cutoff = 3
    kappa_P_RTA, kappa_C = _get_kappa_RTA(aln_lda, [7, 7, 5])
    np.testing.assert_allclose(aln_lda_kappa_P_RTA_with_sigmas, kappa_P_RTA, atol=0.5)
    np.testing.assert_allclose(aln_lda_kappa_C_with_sigmas, kappa_C, atol=0.02)
    aln_lda.sigmas = None
    aln_lda.sigma_cutoff = None


def _get_kappa_RTA(ph3, mesh, is_isotope=False, is_full_pp=False):
    ph3.mesh_numbers = mesh
    ph3.init_phph_interaction()
    ph3.run_thermal_conductivity(
        temperatures=[
            300,
        ],
        is_isotope=is_isotope,
        is_full_pp=is_full_pp,
        conductivity_type="wigner",
    )
    return (
        ph3.thermal_conductivity.kappa_P_RTA.ravel(),
        ph3.thermal_conductivity.kappa_C.ravel(),
    )



"""Tests of Phono3py API."""
from pathlib import Path

from phono3py import Phono3py

cwd = Path(__file__).parent


def test_displacements_setter_NaCl(nacl_pbe: Phono3py):
    """Test Phono3py.displacements setter.

    Just check no error.

    """
    ph3_in = nacl_pbe
    displacements = ph3_in.displacements
    ph3 = Phono3py(
        ph3_in.unitcell,
        supercell_matrix=ph3_in.supercell_matrix,
        primitive_matrix=ph3_in.primitive_matrix,
    )
    ph3.displacements = displacements


def test_displacements_setter_Si(si_pbesol_111_222_fd: Phono3py):
    """Test Phono3py.displacements setter and Phono3py.phonon_displacements setter.

    Just check no error.

    """
    ph3_in = si_pbesol_111_222_fd
    displacements = ph3_in.displacements
    phonon_displacements = ph3_in.phonon_displacements
    ph3 = Phono3py(
        ph3_in.unitcell,
        supercell_matrix=ph3_in.supercell_matrix,
        phonon_supercell_matrix=ph3_in.phonon_supercell_matrix,
        primitive_matrix=ph3_in.primitive_matrix,
    )
    ph3.displacements = displacements
    ph3.phonon_displacements = phonon_displacements
